import gc

from axiom.test.historic.stubloader import StubbedTest
from axiom.errors import ItemNotFound
from axiom.item import Item
from axiom.attributes import text
from axiom.upgrade import (
    registerAttributeCopyingUpgrader, upgradeExplicitOid, _hasExplicitOid)


class Dummy(Item):
    typeName = 'axiom_storeid_dummy'
    schemaVersion = 2

    attribute = text(doc='text', allowNone=False)


registerAttributeCopyingUpgrader(Dummy, 1, 2)


class Dummy2(Item):
    typeName = 'axiom_storeid_dummy2'
    schemaVersion = 1

    attribute = text(doc='text', allowNone=False)



class StoreIDTransitionTest(StubbedTest):
    def test_transition(self):
        """
        Test that the items survived the transition to explicit oids.
        """
        # Make sure we push the upgraded items out of cache
        gc.collect()

        self.assertEquals(self.store.getItemByID(1).attribute, u'one')
        self.assertEquals(
            self.store.findUnique(Dummy, Dummy.attribute == u'two').storeID,
            2)
        self.assertRaises(ItemNotFound, self.store.getItemByID, 3)
        i2 = self.store.getItemByID(4)
        self.assertEquals(i2.attribute, u'four')
        self.assertIsInstance(i2, Dummy2)


    def test_vacuum(self):
        """
        Test that the items survive vacuuming after upgrading.
        """
        # This will upgrade Dummy2 in-place; Dummy had its own upgrade.
        upgradeExplicitOid(self.store)

        # Make sure we push the upgraded items out of cache
        gc.collect()
        self.store.executeSQL('VACUUM')
        self.test_transition()


    def test_upgradeOid(self):
        """
        When an item type is upgraded, the new table has an explicit oid
        column.
        """
        self.assertTrue(
            _hasExplicitOid(self.store, 'item_axiom_storeid_dummy_v2'))
        self.assertFalse(
            _hasExplicitOid(self.store, 'item_axiom_storeid_dummy2_v1'))


    def test_oids(self):
        """
        Explicit oid columns are added to tables by L{upgradeExplicitOid}.
        """
        upgradeExplicitOid(self.store)
        self.assertTrue(_hasExplicitOid(self.store, 'axiom_objects'))
        self.assertTrue(_hasExplicitOid(self.store, 'axiom_types'))
        self.assertTrue(
            _hasExplicitOid(self.store, 'item_axiom_storeid_dummy_v2'))
        self.assertTrue(
            _hasExplicitOid(self.store, 'item_axiom_storeid_dummy2_v1'))



# -*- test-case-name: axiom.test.test_batch -*-

"""
Utilities for performing repetitive tasks over potentially large sets
of data over an extended period of time.
"""

import weakref, datetime, os, sys

from zope.interface import implements

from twisted.python import reflect, failure, log, procutils, util, runtime
from twisted.internet import task, defer, reactor, error, protocol
from twisted.application import service

from epsilon import extime, process, cooperator, modal, juice

from axiom import iaxiom, errors as eaxiom, item, attributes
from axiom.scheduler import Scheduler, SubScheduler
from axiom.upgrade import registerUpgrader, registerDeletionUpgrader
from axiom.dependency import installOn

VERBOSE = False

_processors = weakref.WeakValueDictionary()


class _NoWorkUnits(Exception):
    """
    Raised by a _ReliableListener's step() method to indicate it
    didn't do anything.
    """



class _ProcessingFailure(Exception):
    """
    Raised when processItem raises any exception.  This is never raised
    directly, but instances of the three subclasses are.
    """
    def __init__(self, reliableListener, workUnit, failure):
        Exception.__init__(self)
        self.reliableListener = reliableListener
        self.workUnit = workUnit
        self.failure = failure

        # Get rid of all references this failure is holding so that it doesn't
        # cause any crazy object leaks.  See also the comment in
        # BatchProcessingService.step's except suite.
        self.failure.cleanFailure()


    def mark(self):
        """
        Mark the unit of work as failed in the database and update the listener
        so as to skip it next time.
        """
        self.reliableListener.lastRun = extime.Time()
        BatchProcessingError(
            store=self.reliableListener.store,
            processor=self.reliableListener.processor,
            listener=self.reliableListener.listener,
            item=self.workUnit,
            error=self.failure.getErrorMessage())



class _ForwardProcessingFailure(_ProcessingFailure):
    """
    An error occurred in a reliable listener while processing items forward
    from the mark.
    """

    def mark(self):
        _ProcessingFailure.mark(self)
        self.reliableListener.forwardMark = self.workUnit.storeID



class _BackwardProcessingFailure(_ProcessingFailure):
    """
    An error occurred in a reliable listener while processing items backwards
    from the mark.
    """
    def mark(self):
        _ProcessingFailure.mark(self)
        self.reliableListener.backwardMark = self.workUnit.storeID



class _TrackedProcessingFailure(_ProcessingFailure):
    """
    An error occurred in a reliable listener while processing items specially
    added to the batch run.
    """



class BatchProcessingError(item.Item):
    processor = attributes.reference(doc="""
    The batch processor which owns this failure.
    """)

    listener = attributes.reference(doc="""
    The listener which caused this error.
    """)

    item = attributes.reference(doc="""
    The item which actually failed to be processed.
    """)

    error = attributes.bytes(doc="""
    The error message which was associated with this failure.
    """)



class _ReliableTracker(item.Item):
    """
    A tracking item for an out-of-sequence item which a reliable listener
    should be given to process.

    These are created when L{_ReliableListener.addItem} is called and the
    specified item is in the range of items which have already been processed.
    """

    processor = attributes.reference(doc="""
    The batch processor which owns this tracker.
    """)

    listener = attributes.reference(doc="""
    The listener which is responsible for this tracker's item.
    """)

    item = attributes.reference(doc="""
    The item which this is tracking.
    """)



class _ReliableListener(item.Item):
    processor = attributes.reference(doc="""
    The batch processor which owns this listener.
    """)

    listener = attributes.reference(doc="""
    The item which is actually the listener.
    """)

    backwardMark = attributes.integer(doc="""
    Store ID of the first Item after the next Item to be processed in
    the backwards direction.  Usually, the Store ID of the Item
    previously processed in the backwards direction.
    """)

    forwardMark = attributes.integer(doc="""
    Store ID of the first Item before the next Item to be processed in
    the forwards direction.  Usually, the Store ID of the Item
    previously processed in the forwards direction.
    """)

    lastRun = attributes.timestamp(doc="""
    Time indicating the last chance given to this listener to do some
    work.
    """)

    style = attributes.integer(doc="""
    Either L{iaxiom.LOCAL} or L{iaxiom.REMOTE}. Indicates where the
    batch processing should occur, in the main process or a
    subprocess.
    """)

    def __repr__(self):
        return '<ReliableListener %s %r #%r>' % ({iaxiom.REMOTE: 'remote',
                                                  iaxiom.LOCAL: 'local'}[self.style],
                                                 self.listener,
                                                 self.storeID)


    def addItem(self, item):
        assert type(item) is self.processor.workUnitType, \
               "Adding work unit of type %r to listener for type %r" % (
            type(item), self.processor.workUnitType)
        if item.storeID >= self.backwardMark and item.storeID <= self.forwardMark:
            _ReliableTracker(store=self.store,
                             listener=self,
                             item=item)


    def _forwardWork(self, workUnitType):
        if VERBOSE:
            log.msg("%r looking forward from %r" % (self, self.forwardMark,))
        return self.store.query(
            workUnitType,
            workUnitType.storeID > self.forwardMark,
            sort=workUnitType.storeID.ascending,
            limit=2)


    def _backwardWork(self, workUnitType):
        if VERBOSE:
            log.msg("%r looking backward from %r" % (self, self.backwardMark,))
        if self.backwardMark == 0:
            return []
        return self.store.query(
            workUnitType,
            workUnitType.storeID < self.backwardMark,
            sort=workUnitType.storeID.descending,
            limit=2)


    def _extraWork(self):
        return self.store.query(_ReliableTracker,
                                _ReliableTracker.listener == self,
                                limit=2)


    def _doOneWork(self, workUnit, failureType):
        if VERBOSE:
            log.msg("Processing a unit of work: %r" % (workUnit,))
        try:
            self.listener.processItem(workUnit)
        except:
            f = failure.Failure()
            if VERBOSE:
                log.msg("Processing failed: %s" % (f.getErrorMessage(),))
                log.err(f)
            raise failureType(self, workUnit, f)


    def step(self):
        first = True
        for workTracker in self._extraWork():
            if first:
                first = False
            else:
                return True
            item = workTracker.item
            workTracker.deleteFromStore()
            self._doOneWork(item, _TrackedProcessingFailure)

        for workUnit in self._forwardWork(self.processor.workUnitType):
            if first:
                first = False
            else:
                return True
            self.forwardMark = workUnit.storeID
            self._doOneWork(workUnit, _ForwardProcessingFailure)

        for workUnit in self._backwardWork(self.processor.workUnitType):
            if first:
                first = False
            else:
                return True
            self.backwardMark = workUnit.storeID
            self._doOneWork(workUnit, _BackwardProcessingFailure)

        if first:
            raise _NoWorkUnits()
        if VERBOSE:
            log.msg("%r.step() returning False" % (self,))
        return False



class _BatchProcessorMixin:

    def step(self, style=iaxiom.LOCAL, skip=()):
        now = extime.Time()
        first = True

        for listener in self.store.query(_ReliableListener,
                                         attributes.AND(_ReliableListener.processor == self,
                                                        _ReliableListener.style == style,
                                                        _ReliableListener.listener.notOneOf(skip)),
                                         sort=_ReliableListener.lastRun.ascending):
            if not first:
                if VERBOSE:
                    log.msg("Found more work to do, returning True from %r.step()" % (self,))
                return True
            listener.lastRun = now
            try:
                if listener.step():
                    if VERBOSE:
                        log.msg("%r.step() reported more work to do, returning True from %r.step()" % (listener, self))
                    return True
            except _NoWorkUnits:
                if VERBOSE:
                    log.msg("%r.step() reported no work units" % (listener,))
            else:
                first = False
        if VERBOSE:
            log.msg("No listeners left with work, returning False from %r.step()" % (self,))
        return False


    def run(self):
        """
        Try to run one unit of work through one listener.  If there are more
        listeners or more work, reschedule this item to be run again in
        C{self.busyInterval} milliseconds, otherwise unschedule it.

        @rtype: L{extime.Time} or C{None}
        @return: The next time at which to run this item, used by the scheduler
        for automatically rescheduling, or None if there is no more work to do.
        """
        now = extime.Time()
        if self.step():
            self.scheduled = now + datetime.timedelta(milliseconds=self.busyInterval)
        else:
            self.scheduled = None
        return self.scheduled


    def timedEventErrorHandler(self, timedEvent, failureObj):
        failureObj.trap(_ProcessingFailure)
        log.msg("Batch processing failure")
        log.err(failureObj.value.failure)
        failureObj.value.mark()
        return extime.Time() + datetime.timedelta(milliseconds=self.busyInterval)


    def addReliableListener(self, listener, style=iaxiom.LOCAL):
        """
        Add the given Item to the set which will be notified of Items
        available for processing.

        Note: Each Item is processed synchronously.  Adding too many
        listeners to a single batch processor will cause the L{step}
        method to block while it sends notification to each listener.

        @param listener: An Item instance which provides a
        C{processItem} method.

        @return: An Item representing L{listener}'s persistent tracking state.
        """
        existing = self.store.findUnique(_ReliableListener,
                                         attributes.AND(_ReliableListener.processor == self,
                                                        _ReliableListener.listener == listener),
                                         default=None)
        if existing is not None:
            return existing

        for work in self.store.query(self.workUnitType,
                                     sort=self.workUnitType.storeID.descending,
                                     limit=1):
            forwardMark = work.storeID
            backwardMark = work.storeID + 1
            break
        else:
            forwardMark = 0
            backwardMark = 0

        if self.scheduled is None:
            self.scheduled = extime.Time()
            iaxiom.IScheduler(self.store).schedule(self, self.scheduled)

        return _ReliableListener(store=self.store,
                                 processor=self,
                                 listener=listener,
                                 forwardMark=forwardMark,
                                 backwardMark=backwardMark,
                                 style=style)


    def removeReliableListener(self, listener):
        """
        Remove a previously added listener.
        """
        self.store.query(_ReliableListener,
                         attributes.AND(_ReliableListener.processor == self,
                                        _ReliableListener.listener == listener)).deleteFromStore()
        self.store.query(BatchProcessingError,
                         attributes.AND(BatchProcessingError.processor == self,
                                        BatchProcessingError.listener == listener)).deleteFromStore()


    def getReliableListeners(self):
        """
        Return an iterable of the listeners which have been added to
        this batch processor.
        """
        for rellist in self.store.query(_ReliableListener, _ReliableListener.processor == self):
            yield rellist.listener


    def getFailedItems(self):
        """
        Return an iterable of two-tuples of listeners which raised an
        exception from C{processItem} and the item which was passed as
        the argument to that method.
        """
        for failed in self.store.query(BatchProcessingError, BatchProcessingError.processor == self):
            yield (failed.listener, failed.item)


    def itemAdded(self):
        """
        Called to indicate that a new item of the type monitored by this batch
        processor is being added to the database.

        If this processor is not already scheduled to run, this will schedule
        it.  It will also start the batch process if it is not yet running and
        there are any registered remote listeners.
        """
        localCount = self.store.query(
            _ReliableListener,
            attributes.AND(_ReliableListener.processor == self,
                           _ReliableListener.style == iaxiom.LOCAL),
            limit=1).count()

        remoteCount = self.store.query(
            _ReliableListener,
            attributes.AND(_ReliableListener.processor == self,
                           _ReliableListener.style == iaxiom.REMOTE),
            limit=1).count()

        if localCount and self.scheduled is None:
            self.scheduled = extime.Time()
            iaxiom.IScheduler(self.store).schedule(self, self.scheduled)
        if remoteCount:
            batchService = iaxiom.IBatchService(self.store, None)
            if batchService is not None:
                batchService.start()



def upgradeProcessor1to2(oldProcessor):
    """
    Batch processors stopped polling at version 2, so they no longer needed the
    idleInterval attribute.  They also gained a scheduled attribute which
    tracks their interaction with the scheduler.  Since they stopped polling,
    we also set them up as a timed event here to make sure that they don't
    silently disappear, never to be seen again: running them with the scheduler
    gives them a chance to figure out what's up and set up whatever other state
    they need to continue to run.

    Since this introduces a new dependency of all batch processors on a powerup
    for the IScheduler, install a Scheduler or a SubScheduler if one is not
    already present.
    """
    newProcessor = oldProcessor.upgradeVersion(
        oldProcessor.typeName, 1, 2,
        busyInterval=oldProcessor.busyInterval)
    newProcessor.scheduled = extime.Time()

    s = newProcessor.store
    sch = iaxiom.IScheduler(s, None)
    if sch is None:
        if s.parent is None:
            # Only site stores have no parents.
            sch = Scheduler(store=s)
        else:
            # Substores get subschedulers.
            sch = SubScheduler(store=s)
        installOn(sch, s)

    # And set it up to run.
    sch.schedule(newProcessor, newProcessor.scheduled)
    return newProcessor

def processor(forType):
    """
    Create an Axiom Item type which is suitable to use as a batch processor for
    the given Axiom Item type.

    Processors created this way depend on a L{iaxiom.IScheduler} powerup on the
    on which store they are installed.

    @type forType: L{item.MetaItem}
    @param forType: The Axiom Item type for which to create a batch processor
    type.

    @rtype: L{item.MetaItem}

    @return: An Axiom Item type suitable for use as a batch processor.  If such
    a type previously existed, it will be returned.  Otherwise, a new type is
    created.
    """
    MILLI = 1000
    try:
        processor = _processors[forType]
    except KeyError:
        def __init__(self, *a, **kw):
            item.Item.__init__(self, *a, **kw)
            self.store.powerUp(self, iaxiom.IBatchProcessor)

        attrs = {
            '__name__': 'Batch_' + forType.__name__,

            '__module__': forType.__module__,

            '__init__': __init__,

            '__repr__': lambda self: '<Batch of %s #%d>' % (reflect.qual(self.workUnitType), self.storeID),

            'schemaVersion': 2,

            'workUnitType': forType,

            'scheduled': attributes.timestamp(doc="""
            The next time at which this processor is scheduled to run.
            """, default=None),

            # MAGIC NUMBERS AREN'T THEY WONDERFUL?
            'busyInterval': attributes.integer(doc="", default=MILLI / 10),
            }
        _processors[forType] = processor = item.MetaItem(
            attrs['__name__'],
            (item.Item, _BatchProcessorMixin),
            attrs)

        registerUpgrader(
            upgradeProcessor1to2,
            _processors[forType].typeName,
            1, 2)

    return processor



class ProcessUnavailable(Exception):
    """Indicates the process is not available to perform tasks.

    This is a transient error.  Calling code should handle it by
    arranging to do the work they planned on doing at a later time.
    """



class Shutdown(juice.Command):
    """
    Abandon, belay, cancel, cease, close, conclude, cut it out, desist,
    determine, discontinue, drop it, end, finish, finish up, give over, go
    amiss, go astray, go wrong, halt, have done with, hold, knock it off, lay
    off, leave off, miscarry, perorate, quit, refrain, relinquish, renounce,
    resolve, scrap, scratch, scrub, stay, stop, terminate, wind up.
    """
    commandName = "Shutdown"
    responseType = juice.QuitBox


def _childProcTerminated(self, err):
    self.mode = 'stopped'
    err = ProcessUnavailable(err)
    for d in self.waitingForProcess:
        d.errback(err)
    del self.waitingForProcess


class ProcessController(object):
    """
    Stateful class which tracks a Juice connection to a child process.

    Communication occurs over stdin and stdout of the child process.  The
    process is launched and restarted as necessary.  Failures due to the child
    process terminating, either unilaterally of by request, are represented as
    a transient exception class,

    Mode is one of::

      - 'stopped'       (no process running or starting)
      - 'starting'      (process begun but not ready for requests)
      - 'ready'         (process ready for requests)
      - 'stopping'      (process being torn down)
      - 'waiting_ready' (process beginning but will be shut down
                         as soon as it starts up)

    Transitions are as follows::

       getProcess:
           stopped -> starting:
               launch process
               create/save in waitingForStartup/return Deferred
           starting -> starting:
               create/save/return Deferred
           ready -> ready:
                return saved process
           stopping:
                return failing Deferred indicating transient failure
           waiting_ready:
                return failing Deferred indicating transient failure

       stopProcess:
           stopped -> stopped:
               return succeeding Deferred
           starting -> waiting_ready:
               create Deferred, add transient failure errback handler, return
           ready -> stopping:
               call shutdown on process
               return Deferred which fires when shutdown is done

       childProcessCreated:
           starting -> ready:
               callback saved Deferreds
               clear saved Deferreds
           waiting_ready:
               errback saved Deferred indicating transient failure
               return _shutdownIndexerProcess()

       childProcessTerminated:
           starting -> stopped:
               errback saved Deferreds indicating transient failure
           waiting_ready -> stopped:
               errback saved Deferreds indicating transient failure
           ready -> stopped:
               drop reference to process object
           stopping -> stopped:
               Callback saved shutdown deferred

    @ivar process: A reference to the process object.  Set in every non-stopped
    mode.

    @ivar juice: A reference to the juice protocol.  Set in all modes.

    @ivar connector: A reference to the process protocol.  Set in every
    non-stopped mode.

    @ivar onProcessStartup: None or a no-argument callable which will
    be invoked whenever the connection is first established to a newly
    spawned child process.

    @ivar onProcessTermination: None or a no-argument callable which
    will be invoked whenever a Juice connection is lost, except in the
    case where process shutdown was explicitly requested via
    stopProcess().
    """

    __metaclass__ = modal.ModalType

    initialMode = 'stopped'
    modeAttribute = 'mode'

    # A reference to the Twisted process object which corresponds to
    # the child process we have spawned.  Set to a non-None value in
    # every state except stopped.
    process = None

    # A reference to the process protocol object via which we
    # communicate with the process's stdin and stdout.  Set to a
    # non-None value in every state except stopped.
    connector = None

    def __init__(self, name, juice, tacPath,
                 onProcessStartup=None,
                 onProcessTermination=None,
                 logPath=None,
                 pidPath=None):
        self.name = name
        self.juice = juice
        self.tacPath = tacPath
        self.onProcessStartup = onProcessStartup
        self.onProcessTermination = onProcessTermination
        if logPath is None:
            logPath = name + '.log'
        if pidPath is None:
            pidPath = name + '.pid'
        self.logPath = logPath
        self.pidPath = pidPath

    def _startProcess(self):
        executable = sys.executable
        env = os.environ

        twistdBinaries = procutils.which("twistd2.4") + procutils.which("twistd")
        if not twistdBinaries:
            return defer.fail(RuntimeError("Couldn't find twistd to start subprocess"))
        twistd = twistdBinaries[0]

        setsid = procutils.which("setsid")

        self.connector = JuiceConnector(self.juice, self)

        args = [
            sys.executable,
            twistd,
            '--logfile=%s' % (self.logPath,)]

        if not runtime.platform.isWindows():
            args.append('--pidfile=%s' % (self.pidPath,))

        args.extend(['-noy',
                     self.tacPath])

        if setsid:
            args = ['setsid'] + args
            executable = setsid[0]

        self.process = process.spawnProcess(
            self.connector, executable, tuple(args), env=env)

    class stopped(modal.mode):
        def getProcess(self):
            self.mode = 'starting'
            self.waitingForProcess = []

            self._startProcess()

            # Mode has changed, this will call some other
            # implementation of getProcess.
            return self.getProcess()

        def stopProcess(self):
            return defer.succeed(None)

    class starting(modal.mode):
        def getProcess(self):
            d = defer.Deferred()
            self.waitingForProcess.append(d)
            return d

        def stopProcess(self):
            def eb(err):
                err.trap(ProcessUnavailable)

            d = defer.Deferred().addErrback(eb)
            self.waitingForProcess.append(d)

            self.mode = 'waiting_ready'
            return d

        def childProcessCreated(self):
            self.mode = 'ready'

            if self.onProcessStartup is not None:
                self.onProcessStartup()

            for d in self.waitingForProcess:
                d.callback(self.juice)
            del self.waitingForProcess

        def childProcessTerminated(self, reason):
            _childProcTerminated(self, reason)
            if self.onProcessTermination is not None:
                self.onProcessTermination()


    class ready(modal.mode):
        def getProcess(self):
            return defer.succeed(self.juice)

        def stopProcess(self):
            self.mode = 'stopping'
            self.onShutdown = defer.Deferred()
            Shutdown().do(self.juice)
            return self.onShutdown

        def childProcessTerminated(self, reason):
            self.mode = 'stopped'
            self.process = self.connector = None
            if self.onProcessTermination is not None:
                self.onProcessTermination()


    class stopping(modal.mode):
        def getProcess(self):
            return defer.fail(ProcessUnavailable("Shutting down"))

        def stopProcess(self):
            return self.onShutdown

        def childProcessTerminated(self, reason):
            self.mode = 'stopped'
            self.process = self.connector = None
            self.onShutdown.callback(None)


    class waiting_ready(modal.mode):
        def getProcess(self):
            return defer.fail(ProcessUnavailable("Shutting down"))

        def childProcessCreated(self):
            # This will put us into the stopped state - no big deal,
            # we are going into the ready state as soon as it returns.
            _childProcTerminated(self, RuntimeError("Shutting down"))

            # Dip into the ready mode for ever so brief an instant so
            # that we can shut ourselves down.
            self.mode = 'ready'
            return self.stopProcess()

        def childProcessTerminated(self, reason):
            _childProcTerminated(self, reason)
            if self.onProcessTermination is not None:
                self.onProcessTermination()



class JuiceConnector(protocol.ProcessProtocol):

    def __init__(self, proto, controller):
        self.juice = proto
        self.controller = controller

    def connectionMade(self):
        log.msg("Subprocess started.")
        self.juice.makeConnection(self)
        self.controller.childProcessCreated()

    # Transport
    disconnecting = False

    def write(self, data):
        self.transport.write(data)

    def writeSequence(self, data):
        self.transport.writeSequence(data)

    def loseConnection(self):
        self.transport.loseConnection()

    def getPeer(self):
        return ('omfg what are you talking about',)

    def getHost(self):
        return ('seriously it is a process this makes no sense',)

    def inConnectionLost(self):
        log.msg("Standard in closed")
        protocol.ProcessProtocol.inConnectionLost(self)

    def outConnectionLost(self):
        log.msg("Standard out closed")
        protocol.ProcessProtocol.outConnectionLost(self)

    def errConnectionLost(self):
        log.msg("Standard err closed")
        protocol.ProcessProtocol.errConnectionLost(self)

    def outReceived(self, data):
        self.juice.dataReceived(data)

    def errReceived(self, data):
        log.msg("Received stderr from subprocess: " + repr(data))

    def processEnded(self, status):
        log.msg("Process ended")
        self.juice.connectionLost(status)
        self.controller.childProcessTerminated(status)



class JuiceChild(juice.Juice):
    """
    Protocol class which runs in the child process

    This just defines one behavior on top of a regular juice protocol: the
    shutdown command, which drops the connection and stops the reactor.
    """
    shutdown = False

    def connectionLost(self, reason):
        juice.Juice.connectionLost(self, reason)
        if self.shutdown:
            reactor.stop()

    def command_SHUTDOWN(self):
        log.msg("Shutdown message received, goodbye.")
        self.shutdown = True
        return {}
    command_SHUTDOWN.command = Shutdown



class SetStore(juice.Command):
    """
    Specify the location of the site store.
    """
    commandName = 'Set-Store'
    arguments = [('storepath', juice.Path())]


class SuspendProcessor(juice.Command):
    """
    Prevent a particular reliable listener from receiving any notifications
    until a L{ResumeProcessor} command is sent or the batch process is
    restarted.
    """
    commandName = 'Suspend-Processor'
    arguments = [('storepath', juice.Path()),
                 ('storeid', juice.Integer())]



class ResumeProcessor(juice.Command):
    """
    Cause a particular reliable listener to begin receiving notifications
    again.
    """
    commandName = 'Resume-Processor'
    arguments = [('storepath', juice.Path()),
                 ('storeid', juice.Integer())]



class CallItemMethod(juice.Command):
    """
    Invoke a particular method of a particular item.
    """
    commandName = 'Call-Item-Method'
    arguments = [('storepath', juice.Path()),
                 ('storeid', juice.Integer()),
                 ('method', juice.String())]


class BatchProcessingControllerService(service.Service):
    """
    Controls starting, stopping, and passing messages to the system process in
    charge of remote batch processing.

    @ivar batchController: A reference to the L{ProcessController} for
        interacting with the batch process, if one exists.  Otherwise C{None}.
    """
    implements(iaxiom.IBatchService)

    batchController = None

    def __init__(self, store):
        self.store = store
        self.setName("Batch Processing Controller")


    def startService(self):
        service.Service.startService(self)
        tacPath = util.sibpath(__file__, "batch.tac")
        proto = BatchProcessingProtocol()
        rundir = self.store.dbdir.child("run")
        logdir = rundir.child("logs")
        for d in rundir, logdir:
            try:
                d.createDirectory()
            except OSError:
                pass
        self.batchController = ProcessController(
            "batch", proto, tacPath,
            self._setStore, self._restartProcess,
            logdir.child("batch.log").path,
            rundir.child("batch.pid").path)


    def _setStore(self):
        return SetStore(storepath=self.store.dbdir).do(self.batchController.juice)


    def _restartProcess(self):
        reactor.callLater(1.0, self.batchController.getProcess)


    def stopService(self):
        service.Service.stopService(self)
        d = self.batchController.stopProcess()
        d.addErrback(lambda err: err.trap(error.ProcessDone))
        return d


    def call(self, itemMethod):
        """
        Invoke the given bound item method in the batch process.

        Return a Deferred which fires when the method has been invoked.
        """
        item = itemMethod.im_self
        method = itemMethod.im_func.func_name
        return self.batchController.getProcess().addCallback(
            CallItemMethod(storepath=item.store.dbdir,
                           storeid=item.storeID,
                           method=method).do)


    def start(self):
        if self.batchController is not None:
            self.batchController.getProcess()


    def suspend(self, storepath, storeID):
        return self.batchController.getProcess().addCallback(
            SuspendProcessor(storepath=storepath, storeid=storeID).do)


    def resume(self, storepath, storeID):
        return self.batchController.getProcess().addCallback(
            ResumeProcessor(storepath=storepath, storeid=storeID).do)



class _SubStoreBatchChannel(object):
    """
    SubStore adapter for passing messages to the batch processing system
    process.

    SubStores are adaptable to L{iaxiom.IBatchService} via this adapter.
    """
    implements(iaxiom.IBatchService)

    def __init__(self, substore):
        self.storepath = substore.dbdir
        self.service = iaxiom.IBatchService(substore.parent)


    def call(self, itemMethod):
        return self.service.call(itemMethod)


    def start(self):
        self.service.start()


    def suspend(self, storeID):
        return self.service.suspend(self.storepath, storeID)


    def resume(self, storeID):
        return self.service.resume(self.storepath, storeID)



def storeBatchServiceSpecialCase(st, pups):
    """
    Adapt a L{Store} to L{IBatchService}.

    If C{st} is a substore, return a simple wrapper that delegates to the site
    store's L{IBatchService} powerup.  Return C{None} if C{st} has no
    L{BatchProcessingControllerService}.
    """
    if st.parent is not None:
        try:
            return _SubStoreBatchChannel(st)
        except TypeError:
            return None
    storeService = service.IService(st)
    try:
        return storeService.getServiceNamed("Batch Processing Controller")
    except KeyError:
        return None



class BatchProcessingProtocol(JuiceChild):
    siteStore = None

    def __init__(self, service=None, issueGreeting=False):
        juice.Juice.__init__(self, issueGreeting)
        self.storepaths = []
        if service is not None:
            service.cooperator = cooperator.Cooperator()
        self.service = service


    def connectionLost(self, reason):
        # In the child process, we are a server.  In the child process, we
        # don't want to keep running after we can't talk to the client anymore.
        if self.isServer:
            reactor.stop()


    def command_SET_STORE(self, storepath):
        from axiom import store

        assert self.siteStore is None

        self.siteStore = store.Store(storepath, debug=False)
        self.subStores = {}
        self.pollCall = task.LoopingCall(self._pollSubStores)
        self.pollCall.start(10.0)

        return {}

    command_SET_STORE.command = SetStore


    def command_SUSPEND_PROCESSOR(self, storepath, storeid):
        return self.subStores[storepath.path].suspend(storeid).addCallback(lambda ign: {})
    command_SUSPEND_PROCESSOR.command = SuspendProcessor


    def command_RESUME_PROCESSOR(self, storepath, storeid):
        return self.subStores[storepath.path].resume(storeid).addCallback(lambda ign: {})
    command_RESUME_PROCESSOR.command = ResumeProcessor


    def command_CALL_ITEM_METHOD(self, storepath, storeid, method):
        return self.subStores[storepath.path].call(storeid, method).addCallback(lambda ign: {})
    command_CALL_ITEM_METHOD.command = CallItemMethod


    def _pollSubStores(self):
        from axiom import store, substore

        # Any service which has encountered an error will have logged it and
        # then stopped.  Prune those here, so that they are noticed as missing
        # below and re-added.
        for path, svc in self.subStores.items():
            if not svc.running:
                del self.subStores[path]

        try:
            paths = set([p.path for p in self.siteStore.query(substore.SubStore).getColumn("storepath")])
        except eaxiom.SQLError as e:
            # Generally, database is locked.
            log.msg("SubStore query failed with SQLError: %r" % (e,))
        except:
            # WTF?
            log.msg("SubStore query failed with bad error:")
            log.err()
        else:
            for removed in set(self.subStores) - paths:
                self.subStores[removed].disownServiceParent()
                del self.subStores[removed]
                if VERBOSE:
                    log.msg("Removed SubStore " + removed)
            for added in paths - set(self.subStores):
                try:
                    s = store.Store(added, debug=False)
                except eaxiom.SQLError as e:
                    # Generally, database is locked.
                    log.msg("Opening sub-Store failed with SQLError: %r" % (e,))
                except:
                    log.msg("Opening sub-Store failed with bad error:")
                    log.err()
                else:
                    self.subStores[added] = BatchProcessingService(s, style=iaxiom.REMOTE)
                    self.subStores[added].setServiceParent(self.service)
                    if VERBOSE:
                        log.msg("Added SubStore " + added)



class BatchProcessingService(service.Service):
    """
    Steps over the L{iaxiom.IBatchProcessor} powerups for a single L{axiom.store.Store}.
    """
    def __init__(self, store, style=iaxiom.LOCAL):
        self.store = store
        self.style = style
        self.suspended = []


    def suspend(self, storeID):
        item = self.store.getItemByID(storeID)
        self.suspended.append(item)
        return item.suspend()


    def resume(self, storeID):
        item = self.store.getItemByID(storeID)
        self.suspended.remove(item)
        return item.resume()


    def call(self, storeID, methodName):
        return defer.maybeDeferred(getattr(self.store.getItemByID(storeID), methodName))


    def items(self):
        return self.store.powerupsFor(iaxiom.IBatchProcessor)


    def processWhileRunning(self):
        """
        Run tasks until stopService is called.
        """
        work = self.step()
        for result, more in work:
            yield result
            if not self.running:
                break
            if more:
                delay = 0.1
            else:
                delay = 10.0
            yield task.deferLater(reactor, delay, lambda: None)


    def step(self):
        while True:
            items = list(self.items())

            if VERBOSE:
                log.msg("Found %d processors for %s" % (len(items), self.store))

            ran = False
            more = False
            while items:
                ran = True
                item = items.pop()
                if VERBOSE:
                    log.msg("Stepping processor %r (suspended is %r)" % (item, self.suspended))
                try:
                    itemHasMore = item.store.transact(item.step, style=self.style, skip=self.suspended)
                except _ProcessingFailure as e:
                    log.msg("%r failed while processing %r:" % (e.reliableListener, e.workUnit))
                    log.err(e.failure)
                    e.mark()

                    # _Fuck_.  /Fuck/.  If user-code in or below (*fuck*)
                    # item.step creates a Failure on any future iteration
                    # (-Fuck-) of this loop, it will get a reference to this
                    # exception instance, since it's in locals and Failures
                    # extract and save locals (Aaarrrrggg).  Get rid of this so
                    # that doesn't happen.  See also the definition of
                    # _ProcessingFailure.__init__.
                    e = None
                else:
                    if itemHasMore:
                        more = True
                yield None, bool(more or items)
            if not ran:
                yield None, more


    def startService(self):
        service.Service.startService(self)
        self.parent.cooperator.coiterate(self.processWhileRunning())


    def stopService(self):
        service.Service.stopService(self)
        self.store.close()



class BatchManholePowerup(item.Item):
    """
    Previously, an L{IConchUser} powerup.  This class is only still defined for
    schema compatibility.  Any instances of it will be deleted by an upgrader.
    See #1001.
    """
    schemaVersion = 2
    unused = attributes.integer(
        doc="Satisfy Axiom requirement for at least one attribute")

registerDeletionUpgrader(BatchManholePowerup, 1, 2)



# -*- test-case-name: axiom.test.historic.test_manhole1to2 -*-
# Copyright 2008 Divmod, Inc.  See LICENSE for details.

from axiom.dependency import installOn
from axiom.batch import BatchManholePowerup
from axiom.test.historic.stubloader import saveStub

def createDatabase(store):
    installOn(BatchManholePowerup(store=store), store)

if __name__ == '__main__':
    saveStub(createDatabase, 16829)




"""
Helpers for writing Axiom tests.
"""

from twisted.python.filepath import FilePath

from twisted.trial.unittest import SkipTest

from axiom.store import Store

_theBaseStorePaths = {}
def _getBaseStorePath(testCase, creator):
    if creator not in _theBaseStorePaths:
        s = creator(testCase)
        _theBaseStorePaths[creator] = s.dbdir
        s.close()
    return _theBaseStorePaths[creator]


def getPristineStore(testCase, creator):
    """
    Get an Axiom Store which has been created and initialized by C{creator} but
    which has been otherwise untouched.  If necessary, C{creator} will be
    called to make one.

    @type testCase: L{twisted.trial.unittest.TestCase}
    @type creator: one-argument callable
    @param creator: A factory for the Store configuration desired.  Will be
    invoked with the testCase instance if necessary.
    @rtype: L{axiom.store.Store}
    """
    dbdir = FilePath(testCase.mktemp())
    basePath = _getBaseStorePath(testCase, creator)
    basePath.copyTo(dbdir)
    return Store(dbdir)



class CommandStubMixin:
    """
    Pretend to be the parent command for a subcommand.
    """
    def getStore(self):
        # fake out "parent" implementation for stuff.
        return self.store


    def getSynopsis(self):
        return '<CommandStubMixin>'

    subCommand = property(lambda self: self.__class__.__name__)



class CommandStub(object):
    """
    Mock for L{axiom.scripts.axiomatic.Options} which is always set as the
    C{parent} attribute of an I{axiomatic} subcommand.

    @ivar _store: The L{Store} associated which will be supplied to the
        subcommand.
    """
    def __init__(self, store, subCommand):
        self._store = store
        self.subCommand = subCommand


    def getSynopsis(self):
        return "Usage: axiomatic [options]"


    def getStore(self):
        return self._store



class QueryCounter:
    """
    This is a counter object which measures the number of VDBE instructions
    SQLite will execute to fulfill a particular query.

    The count of VDBE instructions is very useful as a proxy for CPU time and
    disk usage, because it (as opposed to CPU time and disk usage) is
    deterministic between runs of a given query regardless of various accidents
    of operating-system latency.

    When creating data for a query involving a limit, start with B{more} Items
    than will be returned by the limited query, not exactly the right number.
    SQLite will do a little bit more work in the case where the limit restricts
    the number of Items returned, and this will cause a test to fail even
    though the performance characteristics being demonstrated are actually
    correct.

    Put another way, if you are testing::

        s.query(MyItem, limit=5)


    You should create six instances of C{MyItem} before the first C{measure}
    call and then create one or more additional instances of C{MyItem} before
    the second C{measure} call.
    """

    def __init__(self, store):
        """
        Create a new query counter and install it on the provided store.

        @param store: an axiom L{Store}.
        """
        self.reset()
        self.store = store

        c = self.store.connection._connection
        # XXX: this only works with the pysqlite backend, even _with_ the hack
        # detection; if we ever care about the apsw backend again, we should
        # probably do something about adding the hack to it, adding this as a
        # public Axiom API, or something.
        sph = getattr(c, "set_progress_handler", None)
        if sph is None:
            raise SkipTest(
                "QueryCounter requires PySQLite 2.4 or newer, or a patch "
                "(see <http://initd.org/tracker/pysqlite/ticket/182>) to "
                "expose the set_progress_handler API.")
        sph(self.progressHandler, 1)

    def progressHandler(self):
        """
        This method will be called internally by SQLite for each bytecode executed.

        It increments a counter.

        @return: 0, aka SQLITE_OK, so that this does not abort the current
        query.
        """
        self.counter += 1
        return 0

    def reset(self):
        """Reset the internal counter to 0.
        """
        self.counter = 0

    def measure(self, f, *a, **k):
        """
        The primary public API of this class, which runs a given function and
        counts the number of bytecodes between its start and finish.

        @return: an integer, the number of VDBE instructions executed.
        """
        save = self.counter
        self.reset()
        try:
            f(*a, **k)
        finally:
            result = self.counter
            self.counter = save
        return result



# test-case-name: axiom.test.historic.test_scheduler1to2

"""
Database creator for the test for the upgrade of Scheduler from version 1 to
version 2.
"""

from axiom.test.historic.stubloader import saveStub

from axiom.scheduler import Scheduler
from axiom.dependency import installOn


def createDatabase(store):
    installOn(Scheduler(store=store), store)


if __name__ == '__main__':
    saveStub(createDatabase, 17606)



# -*- test-case-name: axiom.test.test_upgrading.SchemaUpgradeTest.testUpgradeWithMissingVersion -*-


from axiom.item import Item
from axiom.attributes import text, integer, reference, inmemory

from axiom.upgrade import registerUpgrader, registerAttributeCopyingUpgrader

class ActivateHelper:
    activated = 0
    def activate(self):
        self.activated += 1

class Adventurer(ActivateHelper, Item):
    typeName = 'test_app_player'
    schemaVersion = 3

    name = text()
    activated = inmemory()

class InventoryEntry(ActivateHelper, Item):
    typeName = 'test_app_inv'
    schemaVersion = 1

    owner = reference()
    owned = reference()

    activated = inmemory()

class Sword(ActivateHelper, Item):
    typeName = 'test_app_sword'
    schemaVersion = 3

    name = text()
    damagePerHit = integer()
    activated = inmemory()

    def owner():
        def get(self):
            return self.store.findUnique(InventoryEntry,
                                         InventoryEntry.owned == self).owner
        return get,

    owner = property(*owner())


def sword2to3(oldsword):
    newsword = oldsword.upgradeVersion('test_app_sword', 2, 3)
    n = oldsword.store.getOldVersionOf('test_app_sword', 2)
    itrbl = oldsword.store.query(n)
    newsword.name = oldsword.name
    newsword.damagePerHit = oldsword.damagePerHit
    invent = InventoryEntry(store=newsword.store,
                            owner=oldsword.owner,
                            owned=newsword)
    return newsword


registerUpgrader(sword2to3, 'test_app_sword', 2, 3)


####### DOUBLE-LEGACY UPGRADE SPECTACULAR !! ###########

# declare legacy class.

from axiom.item import declareLegacyItem


declareLegacyItem(
    typeName='test_app_player',
    schemaVersion=2,
    attributes=dict(
        name=text(allowNone=True)))

registerAttributeCopyingUpgrader(Adventurer, 2, 3)

declareLegacyItem(typeName = 'test_app_sword',
                  schemaVersion = 2,

                  attributes = dict(name=text(),
                                    damagePerHit=integer(),
                                    owner=reference(),
                                    activated=inmemory()))


def upgradePlayerAndSword(oldplayer):
    newplayer = oldplayer.upgradeVersion('test_app_player', 1, 2)
    newplayer.name = oldplayer.name

    oldsword = oldplayer.sword

    newsword = oldsword.upgradeVersion('test_app_sword', 1, 2,
                                       name=oldsword.name,
                                       damagePerHit=oldsword.hurtfulness * 2,
                                       owner=newplayer)

    return newplayer, newsword

def player1to2(oldplayer):
    newplayer, newsword = upgradePlayerAndSword(oldplayer)
    return newplayer

def sword1to2(oldsword):
    oldPlayerType = oldsword.store.getOldVersionOf('test_app_player', 1)
    oldplayer = list(oldsword.store.query(oldPlayerType,
                                          oldPlayerType.sword == oldsword))[0]
    newplayer, newsword = upgradePlayerAndSword(oldplayer)
    return newsword


registerUpgrader(sword1to2, 'test_app_sword', 1, 2)
registerUpgrader(player1to2, 'test_app_player', 1, 2)





import os

from twisted.trial import unittest
from twisted.python import filepath

from axiom.store import Store

from axiom.item import Item
from axiom.attributes import path

class PathTesterItem(Item):
    schemaVersion = 1
    typeName = 'test_path_thing'

    relpath = path()
    abspath = path(relative=False)



class InStoreFilesTest(unittest.TestCase):
    """
    Tests for files managed by the store.
    """
    def _testFile(self, s):
        """
        Shared part of file creation tests.
        """
        f = s.newFile('test', 'whatever.txt')
        f.write('crap')
        def cb(fpath):
            self.assertEquals(fpath.open().read(), 'crap')

        return f.close().addCallback(cb)


    def test_createFile(self):
        """
        Ensure that file creation works for on-disk stores.
        """
        s = Store(filepath.FilePath(self.mktemp()))
        return self._testFile(s)


    def test_createFileInMemory(self):
        """
        Ensure that file creation works for in-memory stores as well.
        """
        s = Store(filesdir=filepath.FilePath(self.mktemp()))
        return self._testFile(s)

    def test_createFileInMemoryAtString(self):
        """
        The 'filesdir' parameter should accept a string as well, for now.
        """
        s = Store(filesdir=self.mktemp())
        return self._testFile(s)


    def test_noFiledir(self):
        """
        File creation should raise an error if the store has no file directory.
        """
        s = Store()
        self.assertRaises(RuntimeError, s.newFile, "test", "whatever.txt")

class PathAttributesTest(unittest.TestCase):
    def testRelocatingPaths(self):
        spath = self.mktemp()
        npath = self.mktemp()
        s = Store(spath)
        rel = s.newFile("test", "123")
        TEST_STR = "test 123"

        def cb(fpath):
            fpath.setContent(TEST_STR)

            PathTesterItem(store=s,
                           relpath=fpath)

            s.close()
            os.rename(spath, npath)
            s2 = Store(npath)
            pti = list(s2.query(PathTesterItem))[0]

            self.assertEquals(pti.relpath.getContent(),
                              TEST_STR)

        return rel.close().addCallback(cb)




"""
Tests for L{axiom.userbase}.
"""

import datetime, StringIO, sys

from zope.interface import Interface, implements
from zope.interface.verify import verifyObject

from twisted.trial import unittest
from twisted.internet.defer import maybeDeferred

from twisted.cred.portal import Portal, IRealm
from twisted.cred.checkers import ICredentialsChecker
from twisted.cred.error import UnauthorizedLogin
from twisted.cred.credentials import (
    IUsernamePassword, IUsernameHashedPassword, UsernamePassword,
    UsernameHashedPassword)

from twisted.python.filepath import FilePath

from epsilon.extime import Time

from axiom.iaxiom import IScheduler
from axiom.store import Store
from axiom.substore import SubStore
from axiom.scheduler import TimedEvent, _SubSchedulerParentHook
from axiom import userbase
from axiom.item import Item
from axiom.attributes import integer
from axiom.scripts import axiomatic
from axiom import errors
from axiom import dependency

class IGarbage(Interface):
    pass

class GarbageProtocolHandler(Item):
    schemaVersion = 1
    typeName = 'test_login_garbage'

    powerupInterfaces = (IGarbage,)
    garbage = integer()

    implements(IGarbage)

SECRET = 'bananas'

class UserBaseTest(unittest.TestCase):
    """
    Tests for L{axiom.userbase} with an on-disk store.
    @ivar store: The C{Store} object for the items tested.
    """
    def setUp(self):
        """
        Set up for testing with an on-disk store.
        """
        self.store = Store(FilePath(self.mktemp()))


    def logInAndCheck(self, username, domain='localhost'):
        """
        Ensure that logging in via cred succeeds based on the accounts
        managed by L{axiom.userbase.LoginSystem}.
        """
        s = self.store
        def _speedup():
            l = userbase.LoginSystem(store=s)
            dependency.installOn(l, s)
            s.checkpoint()
            p = Portal(IRealm(s),
                       [ICredentialsChecker(s)])

            a = l.addAccount(username, 'localhost', SECRET)
            gph = GarbageProtocolHandler(store=a.avatars.open(),
                                         garbage=0)
            dependency.installOn(gph, gph.store)
            return p, gph

        p, gph = s.transact(_speedup)

        def wasItGph((interface, avatar, logout)):
            self.assertEquals(interface, IGarbage)
            self.assertEquals(avatar, gph)
            logout()

        return p.login(UsernamePassword('bob@localhost', SECRET), None, IGarbage
                       ).addCallback(wasItGph)

    def testBasicLogin(self):
        self.logInAndCheck('bob')

    def testUppercaseLogin(self):
        self.logInAndCheck('BOB')

    def testMixedCaseLogin(self):
        self.logInAndCheck('BoB')


class MemoryUserBaseTest(UserBaseTest):
    """
    Tests for L{axiom.userbase} with an in-memory store.
    @ivar store: The C{Store} object for the items tested.
    """
    def setUp(self):
        """
        Set up for testing with an in-memory store.
        """
        self.store = Store()


class CommandTestCase(unittest.TestCase):
    """
    Integration tests for the 'axiomatic userbase' command.
    """

    def setUp(self):
        self.dbdir = FilePath(self.mktemp())
        self.store = Store(self.dbdir)


    def tearDown(self):
        self.store.close()


    def _login(self, avatarId, password):
        cc = ICredentialsChecker(self.store)
        p = Portal(IRealm(self.store), [cc])
        return p.login(UsernamePassword(avatarId, password), None,
                       lambda orig, default: orig)


    def assertImplements(self, obj, interface):
        """
        Assert that C{obj} can be adapted to C{interface}.

        @param obj: Any Python object.
        @param interface: A L{zope.interface.Interface} that C{obj} should
        implement.
        """
        self.failUnless(interface.providedBy(interface(obj, None)))


    def userbase(self, *args):
        """
        Run 'axiomatic userbase' with the given arguments on database at
        C{dbdir}.

        @return: A list of lines printed to stdout by the axiomatic command.
        """
        output = StringIO.StringIO()
        sys.stdout, stdout = output, sys.stdout
        try:
            axiomatic.main(['-d', self.dbdir.path, 'userbase'] + list(args))
        finally:
            sys.stdout = stdout
        return output.getvalue().splitlines()


    def test_install(self):
        """
        Create a database, install userbase and check that the store
        implements L{IRealm} and L{ICredentialsChecker}. i.e. that userbase
        has been installed. This is an integration test.
        """
        self.userbase('install')
        self.assertImplements(self.store, IRealm)
        self.assertImplements(self.store, ICredentialsChecker)


    def test_userCreation(self):
        """
        Create a user on a store, implicitly installing userbase, then try to
        log in with the user. This is an integration test.
        """
        self.userbase('create', 'alice', 'localhost', SECRET)

        def cb((interface, avatar, logout)):
            ss = avatar.avatars.open()
            self.assertEquals(list(userbase.getAccountNames(ss)),
                              [(u'alice', u'localhost')])
            self.assertEquals(avatar.password, SECRET)
            logout()

        d = self._login('alice@localhost', SECRET)
        return d.addCallback(cb)


    def test_listOnClean(self):
        """
        Check that we are given friendly and informative output when we use
        'userbase list' on a fresh store.
        """
        output = self.userbase('list')
        self.assertEquals(output, ['No accounts'])


    def test_list(self):
        """
        When users exist, 'userbase list' should print their IDs one to a
        line.
        """
        self.userbase('create', 'alice', 'localhost', SECRET)
        self.userbase('create', 'bob', 'localhost', SECRET)
        output = self.userbase('list')
        self.assertEquals(output, ['alice@localhost', 'bob@localhost'])


    def test_listWithDisabled(self):
        """
        Check that '[DISABLED]' is printed after the ID of users with disabled
        accounts.
        """
        self.userbase('create', 'alice', 'localhost', SECRET)
        self.userbase('create', 'bob', 'localhost', SECRET)

        def cb((interface, avatar, logout)):
            avatar.disabled = 1
            output = self.userbase('list')
            self.assertEquals(output,
                              ['alice@localhost', 'bob@localhost [DISABLED]'])

        return self._login('bob@localhost', SECRET).addCallback(cb)


    def test_listOffering(self):
        """
        Mantissa offerings are added as users with a 'username' but no domain.
        Check that the 'list' command prints these correctly.
        """
        name = 'offering-name'
        self.userbase('install')
        realm = IRealm(self.store)
        substoreItem = SubStore.createNew(self.store, ('app', name))
        realm.addAccount(name, None, None, internal=True,
                         avatars=substoreItem)
        output = self.userbase('list')
        self.assertEquals(output, [name])



def pvals(m):
    d = m.persistentValues()
    d.pop('account')
    return d


class AccountTestCase(unittest.TestCase):
    def testAccountNames(self):
        dbdir = FilePath(self.mktemp())
        s = Store(dbdir)
        ls = userbase.LoginSystem(store=s)
        dependency.installOn(ls, s)
        acc = ls.addAccount('username', 'dom.ain', 'password')
        ss = acc.avatars.open()

        self.assertEquals(
            list(userbase.getAccountNames(ss)),
            [('username', 'dom.ain')])

        acc.addLoginMethod(u'nameuser', u'ain.dom')

        names = list(userbase.getAccountNames(ss))
        names.sort()
        self.assertEquals(
            names,
            [('nameuser', 'ain.dom'), ('username', 'dom.ain')])

    def testGetLoginMethods(self):
        """
        Test L{userbase.getLoginMethods}
        """
        dbdir = FilePath(self.mktemp())
        s = Store(dbdir)
        ls = userbase.LoginSystem(store=s)
        dependency.installOn(ls, s)

        acc = ls.addAccount('username', 'dom.ain', 'password', protocol=u'speech')
        ss = acc.avatars.open()

        for protocol in (None, u'speech'):
            self.assertEquals(list(userbase.getAccountNames(ss, protocol)),
                              [('username', 'dom.ain')])

        # defaults to ANY_PROTOCOL
        acc.addLoginMethod(u'username2', u'dom.ain')

        # check that searching for protocol=speech also gives us the
        # ANY_PROTOCOL LoginMethod
        for protocol in (None, u'speech'):
            self.assertEquals(sorted(userbase.getAccountNames(ss, protocol)),
                              [('username', 'dom.ain'),
                               ('username2', 'dom.ain')])


    def testAvatarStoreState(self):
        """
        You can only pass an 'avatars' argument if it doesn't already have an
        account in it.  Some accounts want to have their stores in slightly odd
        places (like offering.py) but you can't have two accounts added which
        both point to the same store.
        """
        dbdir = FilePath(self.mktemp())
        s = Store(dbdir)
        ls = userbase.LoginSystem(store=s)
        dependency.installOn(ls, s)
        acc = ls.addAccount('alice', 'dom.ain', 'password')

        # this is allowed, if weird
        unrelatedAccount = ls.addAccount(
            'elseice', 'dom.ain', 'password',
            avatars=SubStore.createNew(s, ('crazy', 'what')))

        # this is not allowed.
        self.assertRaises(errors.DuplicateUniqueItem,
                          ls.addAccount,
                          'bob', 'ain.dom', 'xpassword',
                          avatars=acc.avatars)

        # Make sure that our stupid call to addAccount did not corrupt
        # anything, because we are stupid
        self.assertEquals(acc.avatars.open().query(userbase.LoginAccount).count(), 1)


    def testParallelLoginMethods(self):
        dbdir = FilePath(self.mktemp())
        s = Store(dbdir)
        ls = userbase.LoginSystem(store=s)
        acc = ls.addAccount(u'username', u'example.com', u'password')
        ss = acc.avatars.open()

        loginMethods = s.query(userbase.LoginMethod)
        subStoreLoginMethods = ss.query(userbase.LoginMethod)

        self.assertEquals(loginMethods.count(), 1)
        self.assertEquals(
            [pvals(m) for m in loginMethods],
            [pvals(m) for m in subStoreLoginMethods])


    def testSiteLoginMethodCreator(self):
        dbdir = FilePath(self.mktemp())
        s = Store(dbdir)
        ls = userbase.LoginSystem(store=s)
        acc = ls.addAccount(u'username', u'example.com', u'password')

        # Do everything twice to make sure repeated calls don't corrupt state
        # somehow
        for i in [0, 1]:
            acc.addLoginMethod(
                localpart=u'anothername',
                domain=u'example.org',
                verified=True,
                protocol=u'test',
                internal=False)

            loginMethods = s.query(
                userbase.LoginMethod, sort=userbase.LoginMethod.storeID.ascending)

            subStoreLoginMethods = acc.avatars.open().query(
                userbase.LoginMethod, sort=userbase.LoginMethod.storeID.ascending)

            self.assertEquals(loginMethods.count(), 2)

            self.assertEquals(
                [pvals(m) for m in loginMethods],
                [pvals(m) for m in subStoreLoginMethods])


    def testUserLoginMethodCreator(self):
        dbdir = FilePath(self.mktemp())
        s = Store(dbdir)
        ls = userbase.LoginSystem(store=s)
        acc = ls.addAccount(u'username', u'example.com', u'password')
        ss = acc.avatars.open()
        subStoreLoginAccount = ss.findUnique(userbase.LoginAccount)

        # Do everything twice to make sure repeated calls don't corrupt state
        # somehow
        for i in [0, 1]:
            subStoreLoginAccount.addLoginMethod(
                localpart=u'anothername',
                domain=u'example.org',
                verified=True,
                protocol=u'test',
                internal=False)

            loginMethods = s.query(
                userbase.LoginMethod, sort=userbase.LoginMethod.storeID.ascending)

            subStoreLoginMethods = ss.query(
                userbase.LoginMethod, sort=userbase.LoginMethod.storeID.ascending)

            self.assertEquals(loginMethods.count(), 2)

            self.assertEquals(
                [pvals(m) for m in loginMethods],
                [pvals(m) for m in subStoreLoginMethods])


    def testDomainNames(self):
        s = Store()
        acc = s
        for localpart, domain, internal in [
            (u'local', u'example.com', True),
            (u'local', u'example.net', True),
            (u'remote', u'example.org', False),
            (u'another', u'example.com', True),
            (u'brokenguy', None, True)]:
            userbase.LoginMethod(
                store=s,
                localpart=localpart,
                domain=domain,
                verified=True,
                account=s,
                protocol=u'test',
                internal=internal)
        self.assertEquals(userbase.getDomainNames(s), [u"example.com", u"example.net"])



class ThingThatMovesAround(Item):
    typeName = 'test_thing_that_moves_around'
    schemaVersion = 1

    superValue = integer()

    def run():
        pass

class SubStoreMigrationTestCase(unittest.TestCase):

    IMPORTANT_VALUE = 159

    localpart = u'testuser'
    domain = u'example.com'

    def setUp(self):
        self.dbdir = FilePath(self.mktemp())
        self.store = Store(self.dbdir)
        self.ls = userbase.LoginSystem(store=self.store)
        self.scheduler = IScheduler(self.store)

        self.account = self.ls.addAccount(
            self.localpart, self.domain, u'PASSWORD')

        self.accountStore = self.account.avatars.open()

        self.ss = IScheduler(self.accountStore)

        self.origdir = self.accountStore.dbdir
        self.destdir = FilePath(self.mktemp())


    def test_extraction(self):
        """
        Ensure that user store extraction works correctly,
        particularly in the presence of timed events.
        """
        thing = ThingThatMovesAround(store=self.accountStore,
                                     superValue=self.IMPORTANT_VALUE)
        self.ss.schedule(thing, Time() + datetime.timedelta(days=1))
        self.test_noTimedEventsExtraction()


    def test_noTimedEventsExtraction(self):
        """
        Ensure that user store extraction works correctly if no timed
        events are present.
        """
        userbase.extractUserStore(self.account, self.destdir)
        self.assertEquals(
            self.ls.accountByAddress(self.localpart, self.domain),
            None)

        self.failIf(list(self.store.query(SubStore, SubStore.storepath == self.origdir)))
        self.origdir.restat(False)
        self.failIf(self.origdir.exists())
        self.failIf(list(self.store.query(_SubSchedulerParentHook)))


    def test_noTimedEventsInsertion(self):
        """
        Test that inserting a user store succeeds if it contains no
        timed events.
        """
        self.test_noTimedEventsExtraction()
        self._testInsertion()


    def test_insertion(self, _deleteDomainDirectory=False):
        """
        Test that inserting a user store succeeds and that the right
        items are placed in the site store as a result.
        """
        self.test_extraction()
        self._testInsertion(_deleteDomainDirectory)
        insertedStore = self.ls.accountByAddress(self.localpart,
                                                 self.domain).avatars.open()
        self.assertEquals(
            insertedStore.findUnique(ThingThatMovesAround).superValue,
            self.IMPORTANT_VALUE)
        siteStoreSubRef = self.store.getItemByID(insertedStore.idInParent)
        ssph = self.store.findUnique(_SubSchedulerParentHook,
                         _SubSchedulerParentHook.subStore == siteStoreSubRef,
                                     default=None)
        self.failUnless(ssph)
        self.failUnless(self.store.findUnique(TimedEvent,
                                              TimedEvent.runnable == ssph))


    def _testInsertion(self, _deleteDomainDirectory=False):
        """
        Helper method for inserting a user store.
        """
        if _deleteDomainDirectory:
            self.store.filesdir.child('account').child(self.domain).remove()

        userbase.insertUserStore(self.store, self.destdir)


    def test_insertionWithNoDomainDirectory(self):
        """
        Test that inserting a user store succeeds even if it is the
        first one in that domain to be inserted.
        """
        self.test_insertion(True)


class RealmTestCase(unittest.TestCase):
    """
    Tests for the L{IRealm} implementation in L{axiom.userbase}.
    """
    localpart = u'testuser'
    domain = u'example.com'
    password = u'password'

    def setUp(self):
        self.store = Store()
        self.realm = userbase.LoginSystem(store=self.store)
        dependency.installOn(self.realm, self.store)


    def test_powerup(self):
        """
        Test that L{LoginSystem} powers up the store for L{IRealm}.
        """
        self.assertIdentical(self.realm, IRealm(self.store))


    def _requestAvatarId(self, credentials):
        return maybeDeferred(self.realm.requestAvatarId, credentials)


    def test_requestNonexistentAvatarId(self):
        """
        Test that trying to authenticate as a user who does not exist fails
        with a L{NoSuchUser} exception.
        """
        username = u'%s@%s' % (self.localpart, self.domain)
        d = self._requestAvatarId(
            UsernamePassword(username, self.password))
        return self.assertFailure(d, errors.NoSuchUser)


    def test_requestMalformedAvatarId(self):
        """
        Test that trying to authenticate as a user without specifying a
        hostname fails with a L{NoSuchUser} exception.
        """
        d = self._requestAvatarId(
            UsernamePassword(self.localpart, self.password))
        return self.assertFailure(d, errors.MissingDomainPart)


    def test_usernamepassword(self):
        """
        L{LoginSystem.requestAvatarId} returns the store identifier of the
        L{LoginAccount} associated with a L{UsernamePassword} credentials
        object if the username and password identify an existing account.
        """
        account = self.realm.addAccount(
            self.localpart, self.domain, self.password)
        username = u'%s@%s' % (self.localpart, self.domain)
        d = self._requestAvatarId(UsernamePassword(username, self.password))
        d.addCallback(self.assertEqual, account.storeID)
        return d


    def test_usernameHashedPasswordDeprecated(self):
        """
        Authenticating with L{twisted.cred.credentials.IUsernameHashedPassword}
        credentials emits a deprecation warning.
        """
        account = self.realm.addAccount(
            self.localpart, self.domain, self.password)
        username = u'%s@%s' % (self.localpart, self.domain)
        aid = self.successResultOf(
            self._requestAvatarId(
                UsernameHashedPassword(username, self.password)))
        self.assertEquals(aid, account.storeID)
        ws = self.flushWarnings()
        self.assertEquals(ws[0]['category'], DeprecationWarning)


    def test_usernamepasswordInvalid(self):
        """
        L{LoginSystem.requestAvatarId} fails with L{UnauthorizedLogin} if
        the password supplied with the L{UsernamePassword} credentials is
        not valid for the provided username.
        """
        account = self.realm.addAccount(
            self.localpart, self.domain, self.password)
        username = u'%s@%s' % (self.localpart, self.domain)
        d = self._requestAvatarId(UsernamePassword(username, u'blahblah'))
        self.assertFailure(d, UnauthorizedLogin)
        return d


    def test_preauthenticated(self):
        """
        L{LoginSystem.requestAvatarId} returns the store identifier of the
        L{LoginAccount} associated with a L{Preauthenticated} credentials
        object.
        """
        account = self.realm.addAccount(
            self.localpart, self.domain, self.password)
        username = u'%s@%s' % (self.localpart, self.domain)
        d = self._requestAvatarId(userbase.Preauthenticated(username))
        d.addCallback(self.assertEqual, account.storeID)
        return d


    def test_setPassword(self):
        """
        L{LoginAccount.setPassword} allows for logging in with the new password
        and not the old.
        """
        account = self.realm.addAccount(
            self.localpart, self.domain, self.password)
        username = u'%s@%s' % (self.localpart, self.domain)
        self.successResultOf(account.setPassword(u'blahblah'))
        self.assertEquals(
            self.successResultOf(
                self._requestAvatarId(
                    UsernamePassword(username, u'blahblah'))),
            account.storeID)
        self.failureResultOf(
            self._requestAvatarId(
                UsernamePassword(username, self.password)),
            UnauthorizedLogin)


    def test_replacePasswordWrong(self):
        """
        L{LoginAccount.replacePassword} fails with L{BadCredentials} if an
        incorrect current password is supplied.
        """
        account = self.realm.addAccount(
            self.localpart, self.domain, self.password)
        self.failureResultOf(
            account.replacePassword(u'blahblah', u'blah'),
            errors.BadCredentials)


    def test_replacePasswordCorrect(self):
        """
        L{LoginAccount.replacePassword} allows for logging in with the new
        password and not the old if the correct current password is supplied.
        """
        account = self.realm.addAccount(
            self.localpart, self.domain, self.password)
        username = u'%s@%s' % (self.localpart, self.domain)
        self.successResultOf(
            account.replacePassword(self.password, u'blahblah'))
        self.assertEquals(
            self.successResultOf(
                self._requestAvatarId(
                    UsernamePassword(username, u'blahblah'))),
            account.storeID)
        self.failureResultOf(
            self._requestAvatarId(
                UsernamePassword(username, self.password)),
            UnauthorizedLogin)



class PreauthenticatedTests(unittest.TestCase):
    """
    Tests for L{userbase.Preauthenticated}.
    """
    def test_repr(self):
        """
        L{userbase.Preauthenticated} has a repr which identifies its type and
        its user.
        """
        self.assertEqual(
            repr(userbase.Preauthenticated(u'foo@bar')),
            '<Preauthenticated: foo@bar>')


    def test_usernamepassword(self):
        """
        L{Preauthenticated} implements L{IUsernamePassword} and succeeds all
        authentication checks.
        """
        creds = userbase.Preauthenticated(u'foo@bar')
        self.assertTrue(
            verifyObject(IUsernamePassword, creds),
            "Preauthenticated does not implement IUsernamePassword")
        self.assertTrue(
            creds.checkPassword('random string'),
            "Preauthenticated did not accept an arbitrary password.")


    def test_usernamehashedpassword(self):
        """
        L{Preauthenticated} implements L{IUsernameHashedPassword} and succeeds
        all authentication checks.
        """
        creds = userbase.Preauthenticated(u'foo@bar')
        self.assertTrue(
            verifyObject(IUsernameHashedPassword, creds),
            "Preauthenticated does not implement IUsernameHashedPassword")
        self.assertTrue(
            creds.checkPassword('arbitrary bytes'),
            "Preauthenticated did not accept an arbitrary password.")





from axiom.item import Item
from axiom.attributes import text, integer, reference

class Player(Item):
    typeName = 'test_app_player'
    schemaVersion = 1

    name = text()
    sword = reference()

class Sword(Item):
    typeName = 'test_app_sword'
    schemaVersion = 1

    name = text()
    hurtfulness = integer()



# Copright 2008 Divmod, Inc.  See LICENSE file for details.
# -*- test-case-name: axiom.test.test_dependency -*-
"""
A dependency management system for items.
"""

import sys, itertools

from zope.interface.advice import addClassAdvisor

from epsilon.structlike import record

from axiom.item import Item
from axiom.attributes import reference, boolean, AND
from axiom.errors import ItemNotFound, DependencyError, UnsatisfiedRequirement

#There is probably a cleaner way to do this.
_globalDependencyMap = {}

def dependentsOf(cls):
    deps = _globalDependencyMap.get(cls, None)
    if deps is None:
        return []
    else:
        return [d[0] for d in deps]

##Totally ripping off z.i

def dependsOn(itemType, itemCustomizer=None, doc='',
              indexed=True, whenDeleted=reference.NULLIFY):
    """
    This function behaves like L{axiom.attributes.reference} but with
    an extra behaviour: when this item is installed (via
    L{axiom.dependency.installOn} on a target item, the
    type named here will be instantiated and installed on the target
    as well.

    For example::

      class Foo(Item):
          counter = integer()
          thingIDependOn = dependsOn(Baz, lambda baz: baz.setup())

    @param itemType: The Item class to instantiate and install.
    @param itemCustomizer: A callable that accepts the item installed
    as a dependency as its first argument. It will be called only if
    an item is created to satisfy this dependency.

    @return: An L{axiom.attributes.reference} instance.
    """

    frame = sys._getframe(1)
    locals = frame.f_locals

    # Try to make sure we were called from a class def.
    if (locals is frame.f_globals) or ('__module__' not in locals):
        raise TypeError("dependsOn can be used only from a class definition.")
    ref = reference(reftype=itemType, doc=doc, indexed=indexed, allowNone=True,
                    whenDeleted=whenDeleted)
    if "__dependsOn_advice_data__" not in locals:
        addClassAdvisor(_dependsOn_advice)
    locals.setdefault('__dependsOn_advice_data__', []).append(
    (itemType, itemCustomizer, ref))
    return ref

def _dependsOn_advice(cls):
    if cls in _globalDependencyMap:
        print "Double advising of %s. dependency map from first time: %s" % (
            cls, _globalDependencyMap[cls])
        #bail if we end up here twice, somehow
        return cls
    for itemType, itemCustomizer, ref in cls.__dict__[
        '__dependsOn_advice_data__']:
        classDependsOn(cls, itemType, itemCustomizer, ref)
    del cls.__dependsOn_advice_data__
    return cls

def classDependsOn(cls, itemType, itemCustomizer, ref):
    _globalDependencyMap.setdefault(cls, []).append(
        (itemType, itemCustomizer, ref))

class _DependencyConnector(Item):
    """
    I am a connector between installed items and their targets.
    """
    installee = reference(doc="The item installed.")
    target = reference(doc="The item installed upon.")
    explicitlyInstalled = boolean(doc="Whether this item was installed"
                                  "explicitly (and thus whether or not it"
                                  "should be automatically uninstalled when"
                                  "nothing depends on it)")


def installOn(self, target):
    """
    Install this object on the target along with any powerup
    interfaces it declares. Also track that the object now depends on
    the target, and the object was explicitly installed (and therefore
    should not be uninstalled by subsequent uninstallation operations
    unless it is explicitly removed).
    """
    _installOn(self, target, True)


def _installOn(self, target, __explicitlyInstalled=False):
    depBlob = _globalDependencyMap.get(self.__class__, [])
    dependencies, itemCustomizers, refs = (map(list, zip(*depBlob))
                                         or ([], [], []))
    #See if any of our dependencies have been installed already
    for dc in self.store.query(_DependencyConnector,
                               _DependencyConnector.target == target):
        if dc.installee.__class__ in dependencies:
            i = dependencies.index(dc.installee.__class__)
            refs[i].__set__(self, dc.installee)
            del dependencies[i], itemCustomizers[i], refs[i]
        if (dc.installee.__class__ == self.__class__
            and self.__class__ in set(
            itertools.chain([blob[0][0] for blob in
                             _globalDependencyMap.values()]))):
            #Somebody got here before we did... let's punt
            raise DependencyError("An instance of %r is already "
                                  "installed on %r." % (self.__class__,
                                                        target))
    #The rest we'll install
    for i, cls in enumerate(dependencies):
        it = cls(store=self.store)
        if itemCustomizers[i] is not None:
            itemCustomizers[i](it)
        _installOn(it, target, False)
        refs[i].__set__(self, it)
    #And now the connector for our own dependency.

    dc = self.store.findUnique(
        _DependencyConnector,
        AND(_DependencyConnector.target==target,
            _DependencyConnector.installee==self,
            _DependencyConnector.explicitlyInstalled==__explicitlyInstalled),
        None)
    assert dc is None, "Dependency connector already exists, wtf are you doing?"
    _DependencyConnector(store=self.store, target=target,
                         installee=self,
                         explicitlyInstalled=__explicitlyInstalled)

    target.powerUp(self)

    callback = getattr(self, "installed", None)
    if callback is not None:
        callback()

def uninstallFrom(self, target):
    """
    Remove this object from the target, as well as any dependencies
    that it automatically installed which were not explicitly
    "pinned" by calling "install", and raising an exception if
    anything still depends on this.
    """

    #did this class powerup on any interfaces? powerdown if so.
    target.powerDown(self)


    for dc in self.store.query(_DependencyConnector,
                               _DependencyConnector.target==target):
        if dc.installee is self:
            dc.deleteFromStore()

    for item in installedUniqueRequirements(self, target):
        uninstallFrom(item, target)

    callback = getattr(self, "uninstalled", None)
    if callback is not None:
        callback()

def installedOn(self):
    """
    If this item is installed on another item, return the install
    target. Otherwise return None.
    """
    try:
        return self.store.findUnique(_DependencyConnector,
                                     _DependencyConnector.installee == self
                                     ).target
    except ItemNotFound:
        return None


def installedDependents(self, target):
    """
    Return an iterable of things installed on the target that
    require this item.
    """
    for dc in self.store.query(_DependencyConnector,
                               _DependencyConnector.target == target):
        depends = dependentsOf(dc.installee.__class__)
        if self.__class__ in depends:
            yield dc.installee

def installedUniqueRequirements(self, target):
    """
    Return an iterable of things installed on the target that this item
    requires and are not required by anything else.
    """

    myDepends = dependentsOf(self.__class__)
    #XXX optimize?
    for dc in self.store.query(_DependencyConnector,
                               _DependencyConnector.target==target):
        if dc.installee is self:
            #we're checking all the others not ourself
            continue
        depends = dependentsOf(dc.installee.__class__)
        if self.__class__ in depends:
            raise DependencyError(
                "%r cannot be uninstalled from %r, "
                "%r still depends on it" % (self, target, dc.installee))

        for cls in myDepends[:]:
            #If one of my dependencies is required by somebody
            #else, leave it alone
            if cls in depends:
                myDepends.remove(cls)

    for dc in self.store.query(_DependencyConnector,
                               _DependencyConnector.target==target):
        if (dc.installee.__class__ in myDepends
            and not dc.explicitlyInstalled):
            yield dc.installee

def installedRequirements(self, target):
    """
    Return an iterable of things installed on the target that this
    item requires.
    """
    myDepends = dependentsOf(self.__class__)
    for dc in self.store.query(_DependencyConnector,
                               _DependencyConnector.target == target):
        if dc.installee.__class__ in myDepends:
            yield dc.installee



def onlyInstallPowerups(self, target):
    """
    Deprecated - L{Item.powerUp} now has this functionality.
    """
    target.powerUp(self)



class requiresFromSite(
    record('powerupInterface defaultFactory siteDefaultFactory',
           defaultFactory=None,
           siteDefaultFactory=None)):
    """
    A read-only descriptor that will return the site store's powerup for a
    given item.

    @ivar powerupInterface: an L{Interface} describing the powerup that the
    site store should be adapted to.

    @ivar defaultFactory: a 1-argument callable that takes the site store and
    returns a value for this descriptor.  This is invoked in cases where the
    site store does not provide a default factory of its own, and this
    descriptor is retrieved from an item in a store with a parent.

    @ivar siteDefaultFactory: a 1-argument callable that takes the site store
    and returns a value for this descriptor.  This is invoked in cases where
    this descriptor is retrieved from an item in a store without a parent.
    """

    def _invokeFactory(self, defaultFactory, siteStore):
        if defaultFactory is None:
            raise UnsatisfiedRequirement()
        return defaultFactory(siteStore)


    def __get__(self, oself, type=None):
        """
        Retrieve the value of this dependency from the site store.
        """
        siteStore = oself.store.parent
        if siteStore is not None:
            pi = self.powerupInterface(siteStore, None)
            if pi is None:
                pi = self._invokeFactory(self.defaultFactory, siteStore)
        else:
            pi = self._invokeFactory(self.siteDefaultFactory, oself.store)
        return pi




# Copyright 2008 Divmod, Inc.  See LICENSE for details

import sys
import os

from twisted.trial import unittest
from twisted.internet import protocol, defer
from twisted.python.util import sibpath
from twisted.python import log, filepath

from epsilon import extime
from axiom import attributes, item, store, errors
from axiom.iaxiom import IStatEvent

from axiom._pysqlite2 import sqlite_version_info


class RevertException(Exception):
    pass


class TestItem(item.Item):
    schemaVersion = 1
    typeName = 'TestItem'
    foo = attributes.integer(indexed=True, default=10)
    bar = attributes.text()
    baz = attributes.timestamp()
    other = attributes.reference()
    booleanT = attributes.boolean()
    booleanF = attributes.boolean()

    activated = attributes.inmemory()
    checkactive = attributes.inmemory()
    checked = attributes.inmemory()

    myStore = attributes.reference()

    attributes.compoundIndex(bar, baz)

    def activate(self):
        self.activated = True
        if getattr(self, 'checkactive', False):
            assert isinstance(self.other, TestItem), repr(self.other)
            assert self.other != self, repr(self.other)
            self.checked = True



class StoreTests(unittest.TestCase):
    def testCreation(self):
        dbdir = filepath.FilePath(self.mktemp())
        s = store.Store(dbdir)
        s.close()


    def testReCreation(self):
        dbdir = filepath.FilePath(self.mktemp())
        s = store.Store(dbdir)
        s.close()
        s = store.Store(dbdir)
        s.close()


    def test_onlyOneDir(self):
        """
        A Store should raise an error if both dbdir and filesdir are specified.
        """
        self.assertRaises(ValueError, store.Store,
                          filepath.FilePath(self.mktemp()), filesdir=filepath.FilePath(self.mktemp()))


    def testTypeToDatabaseNames(self):
        # The real purpose of this test is to have the new get*Name
        # methods explicitely called somewhere in the test suite. The
        # effect itself does not actually matter much. These functions
        # are proven right by the fact that item creation, querying
        # and update are working.

        # I think the following should be ok for anything that vaguely
        # ressembles SQL.

        s = store.Store()
        tn = s.getTableName(TestItem)

        assert tn.startswith(s.databaseName)

        cn = s.getColumnName(TestItem.foo)
        scn = s.getShortColumnName(TestItem.foo)

        assert len(tn) < len(cn)
        assert len(scn) < len(cn)
        assert cn.endswith(scn)
        assert cn.startswith(tn)

        icn = s.getColumnName(TestItem.storeID)
        sicn = s.getShortColumnName(TestItem.storeID)

        assert len(tn) < len(icn)
        assert len(sicn) < len(icn)
        assert icn.endswith(sicn)
        assert icn.startswith(tn)

    def testGetTableName(self):
        """
        Item instances were getting into the table name cache.
        Make sure only classes are accepted.
        """
        s = store.Store()
        self.assertRaises(errors.ItemClassesOnly, s.getTableName, TestItem(store=s))


    def testTableNameCacheDoesntGrow(self):
        """
        Make sure the table name cache doesn't grow out of control anymore.
        """
        s = store.Store()
        tn = s.getTableName(TestItem)
        x = len(s.typeToTableNameCache)
        for i in range(10):
            s.getTableName(TestItem)
        self.assertEquals(x, len(s.typeToTableNameCache))


    def testStoreIDComparerIdentity(self):
        # We really want this to hold, because the storeID object is
        # used like a regular attribute as a key for various caching
        # within store.
        a0 = TestItem.storeID
        a1 = TestItem.storeID
        self.assertIdentical(a0, a1)


    def test_loadTypeSchema(self):
        """
        L{Store._loadTypeSchema} returns a C{dict} mapping item type names and
        versions to a list of tuples giving information about the on-disk
        schema information for each attribute of that version of that type.
        """
        s = store.Store()
        TestItem(store=s)

        loadedSchema = s._loadTypeSchema()
        self.assertEqual(
            loadedSchema[(TestItem.typeName, TestItem.schemaVersion)],
            [('bar', 'TEXT COLLATE NOCASE', False, attributes.text, ''),
             ('baz', 'INTEGER', False, attributes.timestamp, ''),
             ('booleanF', 'BOOLEAN', False, attributes.boolean, ''),
             ('booleanT', 'BOOLEAN', False, attributes.boolean, ''),
             ('foo', 'INTEGER', True, attributes.integer, ''),
             ('myStore', 'INTEGER', True, attributes.reference, ''),
             ('other', 'INTEGER', True, attributes.reference, ''),
             ])


    def test_checkInconsistentTypeSchema(self):
        """
        L{Store._checkTypeSchemaConsistency} raises L{RuntimeError} if the in
        memory schema of the type passed to it differs from the schema stored
        in the database for that type, either by including too few attributes,
        too many attributes, or the wrong type for one of the attributes.
        """
        s = store.Store()

        schema = sorted(
            [(name, attr.sqltype, attr.indexed, attr, attr.doc)
             for (name, attr) in TestItem.getSchema()],
            key=lambda a: a[0])

        # Test a missing attribute
        e1 = self.assertRaises(
            RuntimeError,
            s._checkTypeSchemaConsistency,
            TestItem,
            {(TestItem.typeName, TestItem.schemaVersion): schema[:-1]})
        self.assertEqual(
            e1[0],
            "Schema mismatch on already-loaded "
            "<class 'axiom.test.test_xatop.TestItem'> <'TestItem'> object "
            "version 1:\n"
            "Only in memory:\n"
            "('other', 'INTEGER')")

        # And an extra attribute
        e2 = self.assertRaises(
            RuntimeError,
            s._checkTypeSchemaConsistency,
            TestItem,
            {(TestItem.typeName, TestItem.schemaVersion):
             schema + [('extra', 'INTEGER')]})
        self.assertEqual(
            e2[0],
            "Schema mismatch on already-loaded "
            "<class 'axiom.test.test_xatop.TestItem'> <'TestItem'> object "
            "version 1:\n"
            "Only on disk:\n"
            "('extra', 'INTEGER')")

        # And the wrong type for one of the attributes
        e3 = self.assertRaises(
            RuntimeError,
            s._checkTypeSchemaConsistency,
            TestItem,
            {(TestItem.typeName, TestItem.schemaVersion):
             [(schema[0][0], 'VARCHAR(64) (this is made up)')] + schema[1:]})
        self.assertEqual(
            e3[0],
            "Schema mismatch on already-loaded "
            "<class 'axiom.test.test_xatop.TestItem'> <'TestItem'> object "
            "version 1:\n"
            "Only on disk:\n"
            "('bar', 'VARCHAR(64) (this is made up)')\n"
            "Only in memory:\n"
            "('bar', 'TEXT COLLATE NOCASE')")


    def test_checkOutdatedTypeSchema(self):
        """
        L{Store._checkTypeSchemaConsistency} raises L{RuntimeError} if the type
        passed to it is the most recent in-memory version of that type and is
        older than the newest on disk schema for that type.
        """
        s = store.Store()

        schema = [
            (name, attr.sqltype, attr.indexed, attr, attr.doc)
            for (name, attr) in TestItem.getSchema()]

        self.assertRaises(
            RuntimeError,
            s._checkTypeSchemaConsistency,
            TestItem,
            {(TestItem.typeName, TestItem.schemaVersion): schema,
             (TestItem.typeName, TestItem.schemaVersion + 1): schema})


    def test_checkConsistencyWhenOpened(self):
        """
        L{Store.__init__} checks the consistency of the schema and raises
        L{RuntimeError} for any inconsistency.
        """
        class SoonToChange(item.Item):
            attribute = attributes.integer()

        dbpath = self.mktemp()
        s = store.Store(dbpath)
        SoonToChange(store=s)
        s.close()

        # Get rid of the cached information about this type
        from axiom.item import _typeNameToMostRecentClass
        del _typeNameToMostRecentClass[SoonToChange.typeName]
        del SoonToChange

        class SoonToChange(item.Item):
            attribute = attributes.boolean()

        self.assertRaises(RuntimeError, store.Store, dbpath)


    def test_createAndLoadExistingIndexes(self):
        """
        L{Store._loadExistingIndexes} returns a C{set} containing the names of
        all of the indexes which exist in the store.
        """
        s = store.Store()
        before = s._loadExistingIndexes()
        TestItem(store=s)
        after = s._loadExistingIndexes()
        created = after - before

        self.assertEqual(
            created,
            set([s._indexNameOf(TestItem, ['foo']),
                 s._indexNameOf(TestItem, ['other']),
                 s._indexNameOf(TestItem, ['myStore']),
                 s._indexNameOf(TestItem, ['bar', 'baz'])]))


    def test_loadExistingAttachedStoreIndexes(self):
        """
        If a store is attached to its parent, L{Store._loadExistingIndexes}
        returns just the indexes which exist in the child store.
        """
        secondaryPath = self.mktemp()
        main = store.Store()
        secondary = store.Store(secondaryPath, parent=main, idInParent=17)

        TestItem(store=secondary)
        before = secondary._loadExistingIndexes()
        secondary.attachToParent()
        after = secondary._loadExistingIndexes()

        self.assertEqual(before, after)


    def test_createAttachedStoreIndexes(self):
        """
        Indexes created by the insertion of the first item of a type into a
        store are created in the database which backs that store even if that
        store is attached to another store backed by a different database.
        """
        secondaryPath = self.mktemp()

        main = store.Store()
        secondary = store.Store(secondaryPath, parent=main, idInParent=23)

        before = secondary._loadExistingIndexes()
        secondary.attachToParent()
        TestItem(store=secondary)

        # Close it to detach from the parent.  Detach from the parent so we can
        # re-open and hopefully avoid accidentally getting any results polluted
        # by the parent store (shouldn't happen, because the above test,
        # test_loadExistingAttachedStoreIndexes, makes sure we can inspect
        # indexes without worrying about an attached parent, but I'm being
        # paranoid).
        secondary.close()
        del secondary

        secondary = store.Store(secondaryPath, parent=main, idInParent=23)
        after = secondary._loadExistingIndexes()

        self.assertEqual(
            after - before,
            set([secondary._indexNameOf(TestItem, ['foo']),
                 secondary._indexNameOf(TestItem, ['other']),
                 secondary._indexNameOf(TestItem, ['myStore']),
                 secondary._indexNameOf(TestItem, ['bar', 'baz'])]))


    def test_loadPythonModuleHint(self):
        """
        If the Python definition of a type found in a Store has not yet been
        loaded, the hint in the I{module} column in type table is loaded.
        """
        # Arbitrary constants used in multiple places and processes.
        typeCount = 3
        magicOffset = 17
        baseModuleName = "axiom_unloaded_module_"

        # Path the temporary new modules will be created in.
        importPath = filepath.FilePath(self.mktemp())
        importPath.makedirs()
        sys.path.insert(0, importPath.path)
        self.addCleanup(sys.path.remove, importPath.path)

        # Path the store will be created at.
        dbdir = filepath.FilePath(self.mktemp())

        # Create some source files, each defining an item type.
        for counter in range(typeCount):
            moduleName = baseModuleName + str(counter)

            # Sanity check - the test can't work if any of the modules is
            # already imported.
            self.assertNotIn(moduleName, sys.modules)

            # Write out the source.
            modulePath = importPath.child(moduleName + ".py")
            modulePath.setContent("""\
from axiom.item import Item
from axiom.attributes import integer

class Unloaded(Item):
    value = integer()
""")

        # In another process, so as not to cause the unloaded modules to be
        # loaded in this process, create a store containing instances of the
        # Unloaded types.
        script = filepath.FilePath(self.mktemp())
        script.setContent("""\
from sys import argv
from twisted.python.reflect import namedAny
from axiom.store import Store

dbdir, typeCount, moduleBase, magicOffset = argv[1:]

s = Store(dbdir)
for i in range(int(typeCount)):
    moduleName = moduleBase + str(i)
    namedAny(moduleName).Unloaded(store=s, value=int(magicOffset) + i)
s.close()
""")

        os.system(" ".join([
                    "PYTHONPATH=%s:$PYTHONPATH" % (importPath.path,),
                    sys.executable, script.path,
                    dbdir.path, str(typeCount), baseModuleName,
                    str(magicOffset)]))

        # Another sanity check.  The modules still better not have been
        # imported in this process.
        for counter in range(typeCount):
            self.assertNotIn(baseModuleName + str(counter), sys.modules)

        # Now open the store here.  This only works if the Store figures out it
        # needs to import the modules defining the types.
        s = store.Store(dbdir.path)

        # And to be sure, find the item and make sure it has the value we
        # expect.
        for counter in range(typeCount):
            Unloaded = sys.modules[baseModuleName + str(counter)].Unloaded
            self.assertEqual(
                s.query(Unloaded,
                        Unloaded.value == magicOffset + counter).count(), 1)

    def test_closing(self):
        """
        Closing a store explicitly closes the cursor and connection that were
        used by the store.
        """
        s = store.Store()
        connection = s.connection
        cursor = s.cursor
        self.assertFalse(connection.closed, 'Connection should be open')
        self.assertFalse(cursor.closed, 'Cursor should be open')
        s.close()
        self.assertTrue(connection.closed, 'Connection should be closed')
        self.assertTrue(cursor.closed, 'Cursor should be closed')


    def test_journalMode(self):
        """
        Passing a journalling mode sets that mode on open.
        """
        dbdir = filepath.FilePath(self.mktemp())
        s = store.Store(dbdir, journalMode=u'MEMORY')
        self.assertEquals(
            s.querySchemaSQL('PRAGMA *DATABASE*.journal_mode'),
            [(u'memory',)])


    def test_journalModeNone(self):
        """
        Passing a journalling mode of C{None} sets no mode.
        """
        dbdir = filepath.FilePath(self.mktemp())
        s = store.Store(dbdir, journalMode=u'WAL')
        s.close()
        s = store.Store(dbdir, journalMode=None)
        self.assertEquals(
            s.querySchemaSQL('PRAGMA *DATABASE*.journal_mode'),
            [(u'wal',)])



class FailurePathTests(unittest.TestCase):

    def testNoCrossStoreRefs(self):
        s1 = store.Store()
        s2 = store.Store()

        t1 = TestItem(store=s1)
        self.assertRaises(errors.NoCrossStoreReferences,
                          TestItem, store=s2, other=t1)

        t2 = TestItem(store=s2)

        self.assertRaises(errors.NoCrossStoreReferences,
                          setattr, t2, 'other', t1)

        self.assertRaises(errors.NoCrossStoreReferences,
                          setattr, t2, 'other', s1)

        t3 = TestItem(other=t1)

        self.assertRaises(errors.NoCrossStoreReferences,
                          setattr, t3, 'store', s2)

        t3.store = s1

        self.assertEquals(list(s1.query(TestItem)),
                          [t1, t3])


class ItemTests(unittest.TestCase):
    def setUp(self):
        self.dbdir = filepath.FilePath(self.mktemp())
        self.store = store.Store(self.dbdir)

    def tearDown(self):
        self.store.close()

    def testFirstActivationHappensWhenAttributesAreSet(self):
        tio = TestItem(store=self.store)
        ti = TestItem(store=self.store,
                      checkactive=True,
                      other=tio)

        self.assertEquals(ti.checked, True)

    def testItemCreation(self):
        timeval = extime.Time.fromISO8601TimeAndDate('2004-10-05T10:12:14.1234')

        s = TestItem(
            foo = 42,
            bar = u'hello world',
            baz = timeval,
            booleanT = True,
            booleanF = False
            )
        s.myStore = self.store

        s.store = self.store
        sid = s.storeID
        self.store.close()
        self.store = store.Store(self.dbdir)
        s2 = self.store.getItemByID(sid)
        self.assertEquals(s2.foo, s.foo)
        self.assertEquals(s2.booleanT, s.booleanT)
        self.assertEquals(s2.booleanF, s.booleanF)
        self.assertIdentical(s2.myStore, self.store)


    def testBasicQuery(self):
        def tt():
            # !@#$ 3x+ speedup over not doing this in a transact()
            created = [TestItem(foo=x, bar=u"string-value-of-"+str(x))
                       for x in range(20)]
            for c in created:
                c.store = self.store

        self.store.transact(tt)

        loaded = self.store.query(TestItem,
                                  TestItem.foo >= 10)

        self.assertEquals(len(list(loaded)), 10)

    def testCreateThenDelete(self):
        timeval = extime.Time.fromISO8601TimeAndDate('2004-10-05T10:12:14.1234')
        sid = []
        def txn():
            s = TestItem(
                store = self.store,
                foo = 42,
                bar = u'hello world',
                baz = timeval,
                booleanT = True,
                booleanF = False
            )
            sid.append(s.storeID)
            self.assertEquals(list(self.store.query(TestItem)), [s])
            s.deleteFromStore()
            self.assertEquals(list(self.store.query(TestItem)), [])
            # hmm.  possibly due its own test.
            # self.assertRaises(KeyError, self.store.getItemByID, sid[0])

        self.store.transact(txn)
        self.assertRaises(KeyError, self.store.getItemByID, sid[0])
        self.assertEquals(list(self.store.query(TestItem)), [])


    def test_getNeverInsertedItem(self):
        """
        Verify that using getItemByID with a default object to attempt to
        load by storeID an Item which was created and deleted within a
        single transaction results in the default object.
        """
        def txn():
            a = TestItem(store=self.store)
            storeID = a.storeID
            a.deleteFromStore()
            del a
            return storeID
        storeID = self.store.transact(txn)
        default = object()
        result = self.store.getItemByID(storeID, default=default)
        self.assertIdentical(result, default)


    def testInMemoryRevert(self):
        item1 = TestItem(
            store=self.store,
            foo=24,
            bar=u'Zoom',
            baz=extime.Time.fromISO8601TimeAndDate('2004-10-05T10:12:14.1234')
            )

        def brokenFunction():
            item2 = TestItem(
                store=self.store,
                foo=42,
                bar=u'mooZ',
                baz=extime.Time.fromISO8601TimeAndDate('1970-03-12T05:05:11.5921')
                )

            item1.foo = 823
            item1.bar = u'this is the wrong answer'
            item1.baz = extime.Time()

            raise RevertException(item2.storeID)

        try:
            self.store.transact(brokenFunction)
        except RevertException as exc:
            [storeID] = exc.args

            self.assertRaises(KeyError, self.store.getItemByID, storeID)
            self.assertEquals(item1.foo, 24)
            self.assertEquals(item1.bar, u'Zoom')
            self.assertEquals(item1.baz.asISO8601TimeAndDate(), '2004-10-05T10:12:14.1234+00:00')
        else:
            self.fail("Transaction should have raised an exception")


    def test_deleteAndVacuum(self):
        """
        VACUUMing the store after deleting an item does not corrupt it.
        """
        sid = []

        @self.store.transact
        def txn1():
            sid.append(
                AttributefulItem(store=self.store, withDefault=0).storeID)
            i = AttributefulItem(store=self.store, withDefault=1)
            sid.append(i.storeID)
            sid.append(
                AttributefulItem(store=self.store, withDefault=2).storeID)
            i.deleteFromStore()

        self.store.executeSQL('VACUUM')

        @self.store.transact
        def txn2():
            self.assertEqual(self.store.getItemByID(sid[0]).withDefault, 0)
            self.assertRaises(
                errors.ItemNotFound, self.store.getItemByID, sid[1])
            self.assertEqual(self.store.getItemByID(sid[2]).withDefault, 2)



class AttributefulItem(item.Item):
    schemaVersion = 1
    typeName = 'test_attributeful_item'

    withDefault = attributes.integer(default=42)
    withoutDefault = attributes.integer()



class StricterItem(item.Item):
    schemaVersion = 1
    typeName = 'test_stricter_item'

    aRef = attributes.reference(allowNone=False)


class AttributeTests(unittest.TestCase):
    def testGetAttribute(self):
        s = store.Store()
        def testGetAttribute():
            x = AttributefulItem(store=s)
            y = AttributefulItem(store=s, withDefault=20)
            z = AttributefulItem(store=s, withoutDefault=30)
            for o in x, y, z:
                o.checkpoint()

            self.assertEquals(x.withDefault, 42)
            self.assertEquals(x.withoutDefault, None)
            self.assertEquals(y.withDefault, 20)
            self.assertEquals(y.withoutDefault, None)
            self.assertEquals(z.withDefault, 42)
            self.assertEquals(z.withoutDefault, 30)
        s.transact(testGetAttribute)

    def testIntegerAttribute_SQLiteBug(self):
        # SQLite 3.2.1 has a bug which causes various integers to be stored
        # incorrect.  For example, 2 ** 48 - 1 is stored as -1.  This is
        # fixed in 3.2.7.

        for power in 8, 16, 24, 32, 48, 63:
            s = store.Store()
            input = 2 ** power - 1
            s.transact(
                AttributefulItem,
                store=s,
                withoutDefault=input)
            output = s.findFirst(AttributefulItem).withoutDefault
            self.assertEquals(input, output)
            s.close()

    if sqlite_version_info < (3, 2, '7'):
        testIntegerAttribute_SQLiteBug.todo = (
            "If this test fails on your system, you should really upgrade SQLite "
            "to at least 3.2.7.  Not doing so will lead to corruption of your "
            "data.")

    def testQueries(self):
        s = store.Store()
        def testQueries():
            x = AttributefulItem(store=s, withDefault=50)
            y = AttributefulItem(store=s, withDefault=30)
            z = AttributefulItem(store=s, withoutDefault=30)

            for o in x, y, z:
                o.checkpoint()

            self.assertEquals(
                list(s.query(AttributefulItem, AttributefulItem.withoutDefault != None,
                             sort=AttributefulItem.withoutDefault.desc)),
                [z])

            self.assertEquals(
                list(s.query(AttributefulItem, sort=AttributefulItem.withDefault.desc)),
                [x, z, y])

        s.transact(testQueries)

    def testDontAllowNone(self):
        s = store.Store()
        def testDontAllowNone():
            try:
                x = StricterItem(store=s)
            except TypeError:
                pass
            else:
                self.fail("Creating a StricterItem without an aRef value should have failed")

            a = AttributefulItem(store=s)
            x = StricterItem(store=s, aRef=a)
            self.assertEquals(x.aRef, a)

            try:
                x.aRef = None
            except TypeError:
                pass
            else:
                self.fail("Setting aRef to None on a StricterItem should have failed")
        s.transact(testDontAllowNone)


class TestFindOrCreate(unittest.TestCase):

    def testCreate(self):
        s = store.Store()
        ai = s.findOrCreate(AttributefulItem)
        self.assertEquals(ai.withDefault, 42)
        self.assertEquals(ai.withoutDefault, None)

    def testFind(self):
        s = store.Store()
        ai = s.findOrCreate(AttributefulItem, withoutDefault=1234)
        ai2 = s.findOrCreate(AttributefulItem, withDefault=42)
        ai3 = s.findOrCreate(AttributefulItem)

        ai4 = s.findOrCreate(AttributefulItem, withDefault=71)
        ai5 = s.findOrCreate(AttributefulItem, withDefault=71)

        self.assertIdentical(ai, ai2)
        self.assertIdentical(ai3, ai2)
        self.assertIdentical(ai4, ai5)
        self.assertNotIdentical(ai, ai4)

    def testIfNew(self):
        l = []
        s = store.Store()

        ai1 = s.findOrCreate(AttributefulItem, l.append, withDefault=1234)
        ai2 = s.findOrCreate(AttributefulItem, l.append, withDefault=1234)
        ai3 = s.findOrCreate(AttributefulItem, l.append, withDefault=4321)
        ai4 = s.findOrCreate(AttributefulItem, l.append, withDefault=4321)

        self.assertEquals(len(l), 2)
        self.assertEquals(l, [ai1, ai3])

    def testFindFirst(self):
        s = store.Store()
        a0 = ai = AttributefulItem(store=s)
        ai2 = s.findFirst(AttributefulItem, AttributefulItem.withDefault == 42)
        shouldBeNone = s.findFirst(AttributefulItem, AttributefulItem.withDefault == 99)
        self.assertEquals(ai, ai2)
        self.assertEquals(shouldBeNone, None)

        ai = AttributefulItem(store=s, withDefault=24)
        ai2 = s.findFirst(AttributefulItem, AttributefulItem.withDefault == 24)
        self.assertEquals(ai, ai2)

        ai = AttributefulItem(store=s, withDefault=55)
        ai2 = s.findFirst(AttributefulItem)
        self.assertEquals(a0, ai2)



class DeletedTrackingItem(item.Item):
    """
    Helper class for testing that C{deleted} is called by
    ItemQuery.deleteFromStore.
    """
    deletedTimes = 0
    value = attributes.integer()

    def deleted(self):
        DeletedTrackingItem.deletedTimes += 1



class DeleteFromStoreTrackingItem(item.Item):
    """
    Helper class for testing that C{deleteFromStore} is called by
    ItemQuery.deleteFromStore.
    """
    deletedTimes = 0
    value = attributes.integer()

    def deleteFromStore(self):
        DeleteFromStoreTrackingItem.deletedTimes += 1
        item.Item.deleteFromStore(self)



class MassInsertDeleteTests(unittest.TestCase):

    def setUp(self):
        self.storepath = filepath.FilePath(self.mktemp())
        self.store = store.Store(self.storepath)

    def testBatchInsert(self):
        """
        Make sure that batchInsert creates all the items it's supposed
        to with appropriate attributes.
        """

        dataRows = [(37, 93),
                    (1,   2)]

        self.store.batchInsert(AttributefulItem,
                              [AttributefulItem.withDefault,
                               AttributefulItem.withoutDefault],
                              dataRows)
        items = list(self.store.query(AttributefulItem))
        self.assertEquals(items[0].withDefault, 37)
        self.assertEquals(items[0].withoutDefault, 93)
        self.assertEquals(items[1].withDefault, 1)
        self.assertEquals(items[1].withoutDefault, 2)

    def testTransactedBatchInsert(self):
        """
        Test that batchInsert works in a transaction.
        """
        dataRows = [(37, 93),
                    (1,   2)]

        self.store.transact(self.store.batchInsert,
                            AttributefulItem,
                            [AttributefulItem.withDefault,
                             AttributefulItem.withoutDefault],
                            dataRows)

        items = list(self.store.query(AttributefulItem))
        self.assertEquals(items[0].withDefault, 37)
        self.assertEquals(items[0].withoutDefault, 93)
        self.assertEquals(items[1].withDefault, 1)
        self.assertEquals(items[1].withoutDefault, 2)

    def testBatchInsertReference(self):
        """
        Test that reference args are handled okay by batchInsert.
        """
        itemA = AttributefulItem(store=self.store)
        itemB = AttributefulItem(store=self.store)
        dataRows = [(1, u"hello", extime.Time(),
                     itemA, True, False, self.store),
                    (2, u"hoorj", extime.Time(),
                     itemB, False, True, self.store)]

        self.store.batchInsert(TestItem,
                               [TestItem.foo, TestItem.bar,
                                TestItem.baz, TestItem.other,
                                TestItem.booleanT, TestItem.booleanF,
                                TestItem.myStore],
                               dataRows)
        items = list(self.store.query(TestItem))

        self.assertEquals(items[0].other, itemA)
        self.assertEquals(items[1].other, itemB)
        self.assertEquals(items[0].store, self.store)
        self.assertEquals(items[1].store, self.store)

    def testMemoryBatchInsert(self):
        """
        Test that batchInsert works on an in-memory store.
        """
        self.store = store.Store()
        self.testBatchInsert()

    def testBatchInsertSelectedAttributes(self):
        """
        Test that batchInsert does the right thing when only a few
        attributes are being set.
        """
        dataRows = [(u"hello", 50, False, self.store),
                    (u"hoorj", None, True, self.store)]

        self.store.batchInsert(TestItem,
                               [TestItem.bar,
                                TestItem.foo,
                                TestItem.booleanF,
                                TestItem.myStore],
                               dataRows)
        items = list(self.store.query(TestItem))

        self.assertEquals(items[0].other, None)
        self.assertEquals(items[1].other, None)
        self.assertEquals(items[0].foo, 50)
        self.assertEquals(items[1].foo, None)
        self.assertEquals(items[0].bar, u"hello")
        self.assertEquals(items[1].bar, u"hoorj")
        self.assertEquals(items[0].store, self.store)
        self.assertEquals(items[1].store, self.store)

    def testBatchDelete(self):
        """
        Ensure that unqualified batchDelete removes all the items of a
        certain class.
        """
        for i in xrange(10):
            AttributefulItem(store=self.store, withoutDefault=i)

        self.store.query(AttributefulItem).deleteFromStore()
        self.assertEquals(list(self.store.query(AttributefulItem)), [])

    def testBatchDeleteCondition(self):
        """
        Ensure that conditions for batchDelete are honored properly.
        """
        for i in xrange(10):
            AttributefulItem(store=self.store, withoutDefault=i)

        self.store.query(AttributefulItem,
                              AttributefulItem.withoutDefault > 4
                         ).deleteFromStore()
        self.assertEquals(self.store.query(AttributefulItem).count(), 5)

    def testSlowBatchDelete(self):
        """
        Ensure that a 'deleted' method on an Item will be called if it exists.
        """
        DeletedTrackingItem(store=self.store)
        self.store.query(DeletedTrackingItem).deleteFromStore()
        self.assertEqual(DeletedTrackingItem.deletedTimes, 1)


    def test_slowBatchDeleteBecauseDeletedFromStore(self):
        """
        Ensure that a 'deleteFromStore' method on an Item will be called if it
        exists.
        """
        DeleteFromStoreTrackingItem(store=self.store)
        self.store.query(DeleteFromStoreTrackingItem).deleteFromStore()
        self.assertEqual(DeleteFromStoreTrackingItem.deletedTimes, 1)


    def test_batchDeleteOrder(self):
        """
        C{deleteFromStore} on a query with an order specified disregards the
        order.
        """
        for i in xrange(10):
            AttributefulItem(store=self.store, withoutDefault=i)
        self.store.query(
            AttributefulItem,
            sort=AttributefulItem.withoutDefault.asc).deleteFromStore()
        self.assertEqual(list(self.store.query(AttributefulItem)), [])


    def test_batchDeleteOrderLimit(self):
        """
        C{deleteFromStore} on a query with an order and limit specified does
        not disregard the order.
        """
        options = self.store.querySQL('PRAGMA compile_options;')
        if (u'ENABLE_UPDATE_DELETE_LIMIT',) not in options:
            raise unittest.SkipTest(
                'SQLite compiled without SQLITE_ENABLE_UPDATE_DELETE_LIMIT')

        for i in xrange(10):
            AttributefulItem(store=self.store, withoutDefault=i)
        self.store.query(
            AttributefulItem,
            sort=AttributefulItem.withoutDefault.desc,
            limit=5).deleteFromStore()
        items = list(self.store.query(
            AttributefulItem, sort=AttributefulItem.withoutDefault.asc))
        self.assertEqual(len(items), 5)
        self.assertEqual(items[-1].withoutDefault, 4)



# Item types we will use to change the underlying database schema (by creating
# them).
class ConcurrentItemA(item.Item):
    anAttribute = attributes.text()

class ConcurrentItemB(item.Item):
    anotherAttribute = attributes.integer()

class ProcessConcurrencyTestCase(unittest.TestCase,
                                 protocol.ProcessProtocol):

    def spawn(self, *args):
        self.d = defer.Deferred()
        from twisted.internet import reactor
        reactor.spawnProcess(
            self,
            sys.executable,
            [sys.executable] + list(args),
            os.environ)
        return self.d

    ok = 0

    def outReceived(self, data):
        if data == '1':
            # step 1: create an item
            cia = ConcurrentItemA(store=self.store,
                                  anAttribute=u'aaa')
            # then tell the subprocess to load it
            self.transport.write(str(cia.storeID)+'\n')
        elif data == '2':
            # step 2: the subprocess has notified us that it has successfully
            # completed
            self.ok = 1

    def errReceived(self, data):
        # we should never *really* get anything to stdout
        print data

    def processEnded(self, reason):
        # total correctness would have us checking the exit code too, but we
        # got all the output we expected, so whatever.
        if self.ok:
            self.d.callback('OK')
        else:
            self.d.errback(reason)

    def testNewItemTypeInSubprocess(self):
        dbdir = filepath.FilePath(self.mktemp())
        self.store = store.Store(dbdir)
        # Open the store and leave its schema empty (don't create any items)
        # until the subprocess has opened it and loaded the bogus schema.
        return self.spawn(sibpath(__file__, "openthenload.py"), dbdir.path)


class ConcurrencyTestCase(unittest.TestCase):
    def testSchemaChange(self):
        """
        When a statement is executed after the underlying schema has been
        changed (tables are added, database is vacuumed, etc) by another thread
        of execution, PySQLite2 will raise an OperationalError.  This is
        basically harmless and the query will work if re-executed.  This should
        be done transparently.
        """
        dbdir = filepath.FilePath(self.mktemp())

        firstStore = store.Store(dbdir)
        ConcurrentItemA(store=firstStore)

        secondStore = store.Store(dbdir)

        self.assertNotIdentical(firstStore, secondStore) # if this line starts
                                                         # breaking, rewrite
                                                         # this test.

        ConcurrentItemB(store=firstStore)

        self.assertEquals(secondStore.query(ConcurrentItemA).count(), 1)


    def testNewItemType(self):
        """
        Creating the first instance of a of an Item subclass changes the
        underlying database schema as well as some Store-private state which
        tracks that schema.  Test to make sure that creating the first instance
        of an Item subclass in one store is reflected in a second store.
        """
        dbdir = filepath.FilePath(self.mktemp())

        firstStore = store.Store(dbdir)
        secondStore = store.Store(dbdir)

        ConcurrentItemA(store=firstStore)
        self.assertEquals(secondStore.query(ConcurrentItemA).count(), 1)



class LoggingTests(unittest.TestCase):
    """
    Tests for log events emitted by L{axiom.store}.
    """
    def _openTest(self, dbdir, expectedValue):
        events = []
        log.addObserver(events.append)
        self.addCleanup(log.removeObserver, events.append)
        store.Store(dbdir).close()
        for event in events:
            if event.get('interface') is not IStatEvent:
                continue
            if event.get('store_opened') != expectedValue:
                continue
            break
        else:
            self.fail("store_opened IStatEvent not emitted")


    def test_openOnDisk(self):
        """
        Opening a file-backed store logs an event including the path to the
        store.
        """
        dbdir = self.mktemp()
        self._openTest(dbdir, os.path.abspath(dbdir))


    def test_openInMemory(self):
        """
        Opening a memory-backed store logs an event with an empty string for
        the path to the store.
        """
        self._openTest(None, '')



# Copyright (c) 2008 Divmod.  See LICENSE for details.

"""
Package for plugins for interfaces in Axiom.
"""

from epsilon.hotfix import require
require('twisted', 'plugin_package_paths')

from twisted.plugin import pluginPackagePaths
__path__.extend(pluginPackagePaths(__name__))
__all__ = []




import random

from twisted.trial.unittest import TestCase

from axiom.store import Store
from axiom.item import Item
from axiom.attributes import integer

from axiom.queryutil import overlapping, AttributeTuple

class Segment(Item):
    typeName = 'test_overlap_segment'
    schemaVersion = 1

    x = integer()
    y = integer()

    def __repr__(self):
        return 'Segment<%d,%d>' % (self.x, self.y)

class ABC(Item):
    typeName = 'test_tuple_queries'
    schemaVersion = 1

    a = integer(allowNone=False)
    b = integer(allowNone=False)
    c = integer(allowNone=False)

class TestQueryUtilities(TestCase):

    def testBetweenQuery(self):
        # From a drawn copy of the docstring:

        s = Store()

        G = 3
        K = 4
        H = C = 5
        A = 8
        D = 11
        E = 17
        B = 20
        F = I = 22
        L = 23
        J = 24

        AB = Segment(store=s, x=A, y=B)
        CD = Segment(store=s, x=C, y=D)
        EF = Segment(store=s, x=E, y=F)
        GH = Segment(store=s, x=G, y=H)
        IJ = Segment(store=s, x=I, y=J)
        KL = Segment(store=s, x=K, y=L)

        AL = Segment(store=s, x=A, y=L)
        CB = Segment(store=s, x=C, y=B)

        CA = Segment(store=s, x=C, y=A)
        BL = Segment(store=s, x=B, y=L)

        self.assertEquals(
            list(s.query(Segment,
                         overlapping(Segment.x,
                                     Segment.y,
                                     A, B),
                         sort=Segment.storeID.asc)),
            [AB, CD, EF, KL, AL, CB, CA, BL],
            )

    ('(((A > 2)) '
     'OR ((A == 2) AND (B > 3)) '
     'OR ((A == 2) AND (B == 3) AND (C >= 4)))')

    def testTupleQueryWithTuples(self):
        s = Store()
        s.transact(self._dotestTupleQueryWithTuples, s)

    def _dotestTupleQueryWithTuples(self, s):
        L = []
        for x in range(3):
            for y in range(3):
                for z in range(3):
                    L.append((x, y, z))
        shuffledL = L[:]
        random.shuffle(shuffledL)
        for a, b, c in shuffledL:
            ABC(a=a,
                b=b,
                c=c,
                store=s)

        at = AttributeTuple(ABC.a, ABC.b, ABC.c)
        for comparee in L:
            qobj = s.query(ABC, at > comparee,
                           sort=[ABC.a.ascending,
                                 ABC.b.ascending,
                                 ABC.c.ascending])
            self.assertEquals(
                L[L.index(comparee) + 1:],
                [(o.a, o.b, o.c) for o in qobj])

        for comparee in L:
            qobj = s.query(ABC, at >= comparee,
                           sort=[ABC.a.ascending,
                                 ABC.b.ascending,
                                 ABC.c.ascending])
            self.assertEquals(
                L[L.index(comparee):],
                [(o.a, o.b, o.c) for o in qobj])

        for comparee in L:
            qobj = s.query(ABC, at == comparee,
                           sort=[ABC.a.ascending,
                                 ABC.b.ascending,
                                 ABC.c.ascending])
            self.assertEquals(
                [comparee],
                [(o.a, o.b, o.c) for o in qobj])

        for comparee in L:
            qobj = s.query(ABC, at != comparee,
                           sort=[ABC.a.ascending,
                                 ABC.b.ascending,
                                 ABC.c.ascending])
            self.assertEquals(
                L[:L.index(comparee)] + L[L.index(comparee) + 1:],
                [(o.a, o.b, o.c) for o in qobj])

        for comparee in L:
            qobj = s.query(ABC, at < comparee,
                           sort=[ABC.a.ascending,
                                 ABC.b.ascending,
                                 ABC.c.ascending])
            self.assertEquals(
                L[:L.index(comparee)],
                [(o.a, o.b, o.c) for o in qobj])

        for comparee in L:
            qobj = s.query(ABC, at <= comparee,
                           sort=[ABC.a.ascending,
                                 ABC.b.ascending,
                                 ABC.c.ascending])
            self.assertEquals(
                L[:L.index(comparee) + 1],
                [(o.a, o.b, o.c) for o in qobj])





from twisted.cred.portal import Portal, IRealm

from twisted.cred.checkers import ICredentialsChecker
from twisted.cred.credentials import UsernamePassword

from axiom.test.test_userbase import IGarbage

from axiom.test.historic import stubloader

CREDENTIALS = (u'test', u'example.com', 'secret')
GARBAGE_LEVEL = 26

class LoginMethodUpgradeTest(stubloader.StubbedTest):
    def testUpgrade(self):
        p = Portal(IRealm(self.store),
                   [ICredentialsChecker(self.store)])

        def loggedIn((interface, avatarAspect, logout)):
            # if we can login, i guess everything is fine
            self.assertEquals(avatarAspect.garbage, GARBAGE_LEVEL)

        creds = UsernamePassword('@'.join(CREDENTIALS[:-1]), CREDENTIALS[-1])
        d = p.login(creds, None, IGarbage)
        return d.addCallback(loggedIn)




from twisted.trial import unittest
from axiom.attributes import integer
from axiom.errors import NoCrossStoreReferences
from axiom.item import Item
from axiom.sequence import List
from axiom.store import Store


class SomeItem(Item):
    schemaVersion = 1
    typeName = 'test_sequence_some_item'
    foo = integer()

    def __repr__(self):
        return '<SomeItem foo=%i at 0x%x>' % (self.foo, id(self))

    def __cmp__(self, other):
        if not isinstance(other, self.__class__):
            return cmp(super(SomeItem, self), other)
        return cmp(self.foo, other.foo)

class SequenceTestCase(unittest.TestCase):
    def setUp(self):
        self.store = Store()
        self.xy = SomeItem(store=self.store, foo=-1)
        for i in range(10):
            item = SomeItem(store=self.store, foo=i)
            setattr(self, 'i%i'%i, item)

    def assertContents(self, seq, L):
        self.assertEquals(len(seq), len(L))
        for i in range(len(L)):
            self.assertIdentical(seq[i], L[i])

class TestSequenceOfItems(SequenceTestCase):
    def test_createItem(self):
        seq = List(store=self.store)
        self.assertEquals(len(seq), 0)

    def test_createItemWithDefaults(self):
        seq = List([self.i0, self.i1], store=self.store)
        self.assertContents(seq, [self.i0, self.i1])

    def test_createItemWithAliens(self):
        otherStore = Store()
        alien1 = SomeItem(store=otherStore, foo=1)
        alien2 = SomeItem(store=otherStore, foo=2)
        alien3 = SomeItem(store=otherStore, foo=3)
        self.assertRaises(NoCrossStoreReferences,
                          List,
                          [alien1, alien2, alien3],
                          store=self.store)

    def test_appendAndGetItem(self):
        seq = List(store=self.store)
        seq.append(self.i0)
        self.assertEquals(len(seq), 1)
        self.assertEquals(seq[0], self.i0)
        seq.append(self.i1)
        seq.append(self.i2)
        self.assertEquals(seq[1], self.i1)
        self.assertEquals(seq[2], self.i2)

    def test_appendSliceSyntax(self):
        seq = List(store=self.store)
        self.assertContents(seq, [])
        seq[len(seq):len(seq)] = [self.i0]
        seq[len(seq):len(seq)] = [self.i1]
        seq[len(seq):len(seq)] = [self.i2]
        self.assertContents(seq, [self.i0,
                                  self.i1,
                                  self.i2])
    test_appendSliceSyntax.todo = "Slices are not supported yet"

    def test_indexErrors(self):
        seq = List(store=self.store)
        self.assertRaises(IndexError, seq.__getitem__, 0)
        seq.append(self.i0)
        self.assertEquals(seq[0], self.i0)
        self.assertRaises(IndexError, seq.__getitem__, 1)

    def test_negativeIndices(self):
        seq = List(store=self.store)
        seq.append(self.i0)
        seq.append(self.i1)
        seq.append(self.i2)
        self.assertEquals(seq[-1], self.i2)
        self.assertEquals(seq[-2], self.i1)
        self.assertEquals(seq[-3], self.i0)
        self.assertRaises(IndexError, seq.__getitem__, -4)

    def test_setItem(self):
        seq = List(store=self.store)
        seq.append(self.i0)
        seq.append(self.i1)
        self.assertEquals(len(seq), 2)
        self.assertEquals(seq[0], self.i0)
        self.assertEquals(seq[1], self.i1)
        seq[1] = self.i2
        self.assertEquals(seq[1], self.i2)

    def test_delItem(self):
        seq = List(store=self.store)
        seq.append(self.i1)
        seq.append(self.i2)
        seq.append(self.i3)
        self.assertEquals(seq.length, 3)
        self.assertEquals(seq[0], self.i1)
        self.assertEquals(seq[1], self.i2)
        self.assertEquals(seq[2], self.i3)
        del seq[1]
        self.assertEquals(seq.length, 2)
        self.assertEquals(seq[0], self.i1)
        self.assertEquals(seq[1], self.i3)
        self.assertRaises(IndexError, seq.__getitem__, 2)


class TestSequenceOperations(SequenceTestCase):
    """
    These test cases were taken from the list of sequence operations
    found at http://docs.python.org/lib/typesseq.html
    """

    def test_x_in_s(self):
        seq = List(store=self.store)
        seq.append(self.i0)
        self.failUnless(self.i0 in seq)
        self.failIf(self.xy in seq)

    def test_x_not_in_s(self):
        seq = List(store=self.store)
        seq.append(self.i0)
        self.failUnless(self.xy not in seq)
        self.failIf(self.i0 not in seq)

    def test_s_plus_t(self):
        L1 = List(store=self.store)
        L2 = List(store=self.store)
        L1.append(self.i0)
        L2.append(self.i1)
        # XXX ASSUMPTION: all operations which return another
        # instance will return regular lists, *not* Lists.
        L = L1 + L2
        self.assertEquals(L, [self.i0, self.i1])

    def test_shallow_copies(self, n=3):
        seq = List(store=self.store)
        seq.append(self.i0)
        for L in (seq * n, n * seq):
            self.assertEquals(L, [self.i0]*n)

    def test_index(self):
        seq = List(store=self.store)
        seq.append(self.i0)
        seq.append(self.i1)
        self.assertIdentical(seq[0], self.i0)
        self.assertIdentical(seq[1], self.i1)
        self.failIfIdentical(seq[0], self.i1)
        self.failIfIdentical(seq[1], self.i0)

    def test_slices(self):
        seq = List(store=self.store)
        seq.append(self.i0)
        seq.append(self.i1)
        seq.append(self.i2)
        seq.append(self.i3)
        self.assertEquals(seq[0:2], [self.i0, self.i1])
        self.assertEquals(seq[0:3], [self.i0, self.i1, self.i2])
        self.assertEquals(seq[1:0], [])
        self.assertEquals(seq[-1:], [self.i3])
    test_slices.todo = "Slices are not supported yet"

    def test_slice_with_step(self):
        seq = List(store=self.store)
        seq.append(self.i0)
        seq.append(self.i1)
        seq.append(self.i2)
        seq.append(self.i3)
        self.assertEquals(seq[0:4:2], [self.i0, self.i2])
        self.assertEquals(seq[1:5:2], [self.i1, self.i3])
    test_slice_with_step.todo = "Slices are not supported yet"

    def test_len(self):
        seq = List(store=self.store)
        self.assertEquals(len(seq), 0)
        seq.append(self.i0)
        self.assertEquals(len(seq), 1)
        seq.append(self.i0)
        self.assertEquals(len(seq), 2)
        seq.append(self.i0)
        self.assertEquals(len(seq), 3)

    def test_min_max(self):
        seq = List(store=self.store)
        seq.append(self.i0)
        seq.append(self.i1)
        seq.append(self.i2)
        self.assertIdentical(min(seq), self.i0)
        self.assertIdentical(max(seq), self.i2)


class TestMutableSequenceOperations(SequenceTestCase):
    """
    These test cases were taken from the list of sequence operations
    found at http://docs.python.org/lib/typesseq-mutable.html

    Some may duplicate L{TestSequenceOperations}, but who cares?
    """

    def test_indexAssignment(self):
        seq = List(store=self.store)
        seq.append(self.i0)
        self.assertIdentical(seq[0], self.i0)
        seq[0] = self.i1
        self.assertIdentical(seq[0], self.i1)

    def test_sliceAssignment(self):
        seq = List(store=self.store)
        seq.append(self.i0)
        seq.append(self.i0)
        seq.append(self.i0)
        seq.append(self.i0)
        self.assertContents(seq, [self.i0,
                                  self.i0,
                                  self.i0,
                                  self.i0])
        seq[1:3] = [self.i1, self.i2]
        self.assertContents(seq, [self.i0,
                                  self.i1,
                                  self.i2,
                                  self.i0])
        seq[1:3] = [self.i3]
        self.assertContents(seq, [self.i0,
                                  self.i3,
                                  self.i0])
        seq[1:3] = []
        self.assertContents(seq, [self.i0])
    test_sliceAssignment.todo = "Slices are not supported yet"

    def test_deleteSlice(self):
        seq = List(store=self.store)
        seq.append(self.i0)
        seq.append(self.i1)
        seq.append(self.i2)
        seq.append(self.i3)
        del seq[1:3]
        self.assertEquals(len(seq), 2)
        self.assertIdentical(seq[0], self.i0)
        self.assertIdentical(seq[1], self.i3)
    test_deleteSlice.todo = "Slices are not supported yet"

    def test_sliceAssignmentStep(self):
        seq = List(store=self.store)
        seq.append(self.i0)
        seq.append(self.i1)
        seq.append(self.i2)
        seq.append(self.i3)
        seq.append(self.i4)
        seq.append(self.i5)
        seq.append(self.i6)
        seq[1:5:2] = [self.i7, self.i7]
        self.assertContents(seq, [self.i0,
                                  self.i7,
                                  self.i2,
                                  self.i7,
                                  self.i4,
                                  self.i5,
                                  self.i6])
    test_sliceAssignmentStep.todo = "Slices are not supported yet"

    def test_deleteSliceStep(self):
        seq = List(store=self.store)
        seq.append(self.i0)
        seq.append(self.i1)
        seq.append(self.i2)
        seq.append(self.i3)
        seq.append(self.i4)
        seq.append(self.i5)
        seq.append(self.i6)
        del seq[1:6:2]
        self.assertContents(seq, [self.i0,
                                  self.i2,
                                  self.i4,
                                  self.i6])
    test_deleteSliceStep.todo = "Slices are not supported yet"

    def test_append(self):
        seq = List(store=self.store)
        self.assertContents(seq, [])
        seq.append(self.i0)
        self.assertContents(seq, [self.i0])
        seq.append(self.i1)
        self.assertContents(seq, [self.i0,
                                  self.i1])

    def test_extend(self):
        L1 = List(store=self.store)
        L1.append(self.i0)
        L1.append(self.i1)
        L2 = List(store=self.store)
        L2.append(self.i2)
        L2.append(self.i3)
        L1.extend(L2)
        self.assertContents(L1, [self.i0,
                                 self.i1,
                                 self.i2,
                                 self.i3])

    def test_extendSliceSyntax(self):
        L1 = List(store=self.store)
        L1.append(self.i0)
        L1.append(self.i1)
        L2 = List(store=self.store)
        L2.append(self.i2)
        L2.append(self.i3)
        L1[len(L1):len(L1)] = L2
        self.assertContents(L1, [self.i0,
                                 self.i1,
                                 self.i2,
                                 self.i3])
    test_extendSliceSyntax.todo = "Slices are not supported yet"

    def test_count(self):
        seq = List(store=self.store)
        seq.append(self.i0)
        seq.append(self.i1)
        seq.append(self.i0)
        seq.append(self.i2)
        seq.append(self.i0)
        seq.append(self.i2)
        self.assertEquals(seq.count(self.i0), 3)
        self.assertEquals(seq.count(self.i1), 1)
        self.assertEquals(seq.count(self.i2), 2)
        self.assertEquals(seq.count(self.i3), 0)

    def test_index(self):
        seq = List(store=self.store)
        seq.append(self.i0)
        seq.append(self.i1)
        seq.append(self.i0)
        seq.append(self.i2)
        seq.append(self.i0)
        seq.append(self.i2)
        self.assertEquals(seq.index(self.i0),    0)
        self.assertEquals(seq.index(self.i0, 0), 0)
        self.assertEquals(seq.index(self.i0, 1), 2)
        self.assertEquals(seq.index(self.i1),    1)
        self.assertEquals(seq.index(self.i1, 1), 1)
        self.assertEquals(seq.index(self.i2),    3)
        self.assertEquals(seq.index(self.i2, 4), 5)
        self.assertRaises(ValueError, seq.index, self.i3)
        self.assertRaises(ValueError, seq.index, self.i1, 3)
        self.assertRaises(ValueError, seq.index, self.i0, 1, 1)
        # TODO: support negative slice boundaries

    def test_insert(self):
        seq = List(store=self.store)
        seq.append(self.i0)
        seq.append(self.i0)
        seq.insert(1, self.i9)
        self.assertContents(seq, [self.i0,
                                  self.i9,
                                  self.i0])

    def test_insertSliceSyntax(self):
        seq = List(store=self.store)
        seq.append(self.i0)
        seq.append(self.i0)
        seq[1:1] = self.i9
        self.assertContents(seq, [self.i0,
                                  self.i9,
                                  self.i0])
    test_insertSliceSyntax.todo = "Slices are not supported yet"

    def test_pop(self):
        seq = List(store=self.store)
        seq.append(self.i0)
        seq.append(self.i1)
        seq.append(self.i2)
        seq.append(self.i3)
        seq.append(self.i4)

        self.assertIdentical(seq.pop(), self.i4)
        self.assertContents(seq, [self.i0,
                                  self.i1,
                                  self.i2,
                                  self.i3])

        self.assertIdentical(seq.pop(0), self.i0)
        self.assertContents(seq, [self.i1,
                                  self.i2,
                                  self.i3])

        self.assertIdentical(seq.pop(-2), self.i2)
        self.assertContents(seq, [self.i1,
                                  self.i3])

        self.assertRaises(IndexError, seq.pop, 13)

    def test_remove(self):
        seq = List(store=self.store)
        seq.append(self.i0)
        seq.append(self.i1)
        seq.append(self.i2)
        seq.append(self.i0)
        seq.append(self.i1)
        seq.append(self.i3)
        seq.append(self.i0)
        self.assertContents(seq, [self.i0,
                                  self.i1,
                                  self.i2,
                                  self.i0,
                                  self.i1,
                                  self.i3,
                                  self.i0])
        seq.remove(self.i0)
        self.assertContents(seq, [self.i1,
                                  self.i2,
                                  self.i0,
                                  self.i1,
                                  self.i3,
                                  self.i0])
        seq.remove(self.i0)
        self.assertContents(seq, [self.i1,
                                  self.i2,
                                  self.i1,
                                  self.i3,
                                  self.i0])
        seq.remove(self.i2)
        self.assertContents(seq, [self.i1,
                                  self.i1,
                                  self.i3,
                                  self.i0])

        self.assertRaises(ValueError, seq.remove, self.i4)

    def test_reverse(self):
        # UPDATE my_list_tbl SET _index = (_index - listlength + 1) * -1
        seq = List(store=self.store)
        seq.append(self.i0)
        seq.append(self.i1)
        seq.append(self.i2)
        seq.append(self.i3)
        self.assertContents(seq, [self.i0,
                                  self.i1,
                                  self.i2,
                                  self.i3])
        seq.reverse()
        self.assertContents(seq, [self.i3,
                                  self.i2,
                                  self.i1,
                                  self.i0])

    '''
    s.sort([cmp[, key[, reverse]]])

    From http://docs.python.org/lib/typesseq-mutable.html

    The sort() method takes optional arguments for controlling the comparisons.

    cmp specifies a custom comparison function of two arguments (list
    items) which should return a negative, zero or positive number
    depending on whether the first argument is considered smaller
    than, equal to, or larger than the second argument:
    "cmp=lambda x,y: cmp(x.lower(), y.lower())"

    key specifies a function of one argument that is used to extract a
    comparison key from each list element:"key=str.lower"

    reverse is a boolean value. If set to True, then the list elements
    are sorted as if each comparison were reversed.

    In general, the key and reverse conversion processes are much
    faster than specifying an equivalent cmp function. This is because
    cmp is called multiple times for each list element while key and
    reverse touch each element only once.

    Changed in version 2.3: Support for None as an equivalent to
    omitting cmp was added.

    Changed in version 2.4: Support for key and reverse was added.
    '''
    def test_sort(self):
        seq = List(store=self.store)
        def seq_randomize():
            while len(seq):
                seq.pop()
            seq.append(self.i3)
            seq.append(self.i0)
            seq.append(self.i1)
            seq.append(self.i4)
            seq.append(self.i2)

        seq_randomize()
        seq.sort()
        self.assertContents(seq, [self.i0,
                                  self.i1,
                                  self.i2,
                                  self.i3,
                                  self.i4])

        seq_randomize()
        seq.sort(lambda x,y: cmp(y,x))
        self.assertContents(seq, [self.i4,
                                  self.i3,
                                  self.i2,
                                  self.i1,
                                  self.i0])

        def strangecmp(x, y):
            xfoo, yfoo = x.foo, y.foo
            if xfoo < 3:
                xfoo += 100
            if yfoo < 3:
                yfoo += 100
            return cmp(xfoo, yfoo)
        seq_randomize()
        seq.sort(strangecmp)
        self.assertContents(seq, [self.i3,
                                  self.i4,
                                  self.i0,
                                  self.i1,
                                  self.i2])

        seq_randomize()
        seq.sort(None, lambda x:x, True)
        self.assertContents(seq, [self.i4,
                                  self.i3,
                                  self.i2,
                                  self.i1,
                                  self.i0])
        seq_randomize()
        seq.sort(strangecmp, lambda x:x, True)
        self.assertContents(seq, [self.i2,
                                  self.i1,
                                  self.i0,
                                  self.i4,
                                  self.i3])

    def test_count(self):
        seq = List(store=self.store)
        seq.append(self.i1)
        seq.append(self.i2)
        seq.append(self.i2)
        seq.append(self.i3)
        seq.append(self.i3)
        seq.append(self.i3)
        self.assertEquals(seq.count(self.i1), 1)
        self.assertEquals(seq.count(self.i2), 2)
        self.assertEquals(seq.count(self.i3), 3)

    def test_contains(self):
        seq = List(store=self.store)
        seq.append(self.i1)
        seq.append(self.i2)
        self.failUnless(self.i1 in seq)
        self.failUnless(self.i2 in seq)
        self.failIf(self.i3 in seq)

    def test_multicontains(self):
        seq1 = List(store=self.store)
        seq2 = List(store=self.store)
        seq1.append(self.i1)
        seq2.append(self.i2)
        self.failUnless(self.i1 in seq1)
        self.failUnless(self.i2 in seq2)
        self.failIf(self.i1 in seq2)
        self.failIf(self.i2 in seq1)

    def test_multidelitem(self):
        seq1 = List(store=self.store)
        seq2 = List(store=self.store)
        seq1.append(self.i1)
        seq1.append(self.i2)
        seq2.append(self.i1)
        seq2.append(self.i2)
        del seq1[0]
        self.assertIdentical(seq2[0], self.i1)
        self.assertIdentical(seq2[1], self.i2)



# -*- test-case-name: axiom.test.test_pysqlite2 -*-

"""
Test code for any cursor implementation which is to work with Axiom.

This probably isn't complete.
"""

from axiom.errors import TimeoutError, TableAlreadyExists, SQLError

class StubCursor(object):
    """
    Stand in for an actual database-backed cursor.  Used by tests to assert the
    right calls are made to execute and to make sure errors from execute are
    handled correctly.

    @ivar statements: A list of SQL strings which have been executed.
    @ivar connection: A reference to the L{StubConnection} which created this
    cursor.
    """
    def __init__(self, connection):
        self.connection = connection
        self.statements = []


    def execute(self, statement, args=()):
        """
        Capture some SQL for later inspection.
        """
        self.statements.append(statement)



class StubConnection(object):
    """
    Stand in for an actual database-backed connection.  Used by tests to create
    L{StubCursors} to easily test behavior of code which interacts with cursors.

    @ivar cursors: A list of all cursors ever created with this connection.
    """
    def __init__(self):
        self.cursors = []


    def cursor(self):
        """
        Create and return a new L{StubCursor}.
        """
        self.cursors.append(StubCursor(self))
        return self.cursors[-1]


    def timeout(self):
        """
        Induce behavior indicative of a database-level transient failure which
        might lead to a timeout.
        """
        raise NotImplementedError



class ConnectionTestCaseMixin:

    # The number of seconds we will allow for timeouts in this test suite.
    TIMEOUT = 5.0

    # The amount of time beyond the specified timeout we will allow Axiom to
    # waste sleeping.  This number shouldn't be changed very often, if ever.
    # We're testing a particular performance feature which we should be able to
    # rely on.
    ALLOWED_SLOP = 0.2


    def createAxiomConnection(self):
        raise NotImplementedError("Cannot create Axiom Connection instance.")


    def createStubConnection(self):
        raise NotImplementedError("Cannot create Axiom Connection instance.")


    def createRealConnection(self):
        """
        Create a memory-backed database connection for integration testing.
        """
        raise NotImplementedError("Real connection creation not implemented.")


    def test_identifyTableCreationError(self):
        """
        When the same table is created twice, we should get a TableAlreadyExists
        exception.
        """
        con = self.createRealConnection()
        cur = con.cursor()
        CREATE_TABLE = "create table foo (bar integer)"
        cur.execute(CREATE_TABLE)
        e = self.assertRaises(TableAlreadyExists, cur.execute, CREATE_TABLE)


    def test_identifyGenericError(self):
        """
        When invalid SQL is issued, we should get a SQLError exception.
        """
        con = self.createRealConnection()
        cur = con.cursor()
        INVALID_STATEMENT = "not an SQL string"
        e = self.assertRaises(SQLError, cur.execute, INVALID_STATEMENT)


    def test_cursor(self):
        """
        Test that the cursor method can actually create a cursor object.
        """
        stubConnection = self.createStubConnection()
        axiomConnection = self.createAxiomConnection(stubConnection)
        axiomCursor = axiomConnection.cursor()

        self.assertEquals(len(stubConnection.cursors), 1)
        statement = "SELECT foo FROM bar"
        axiomCursor.execute(statement)
        self.assertEquals(len(stubConnection.cursors[0].statements), 1)
        self.assertEquals(stubConnection.cursors[0].statements[0], statement)


    def test_timeoutExceeded(self):
        """
        Test that the timeout we pass to the Connection is respected.
        """
        clock = [0]
        def time():
            return clock[0]
        def sleep(n):
            clock[0] += n

        stubConnection = self.createStubConnection()
        axiomConnection = self.createAxiomConnection(stubConnection, timeout=self.TIMEOUT)
        axiomCursor = axiomConnection.cursor()

        axiomCursor.time = time
        axiomCursor.sleep = sleep

        def execute(statement, args=()):
            if time() < self.TIMEOUT * 2:
                return stubConnection.timeout()
            return object()

        stubConnection.cursors[0].execute = execute

        statement = 'SELECT foo FROM bar'
        timeoutException = self.assertRaises(
            TimeoutError,
            axiomCursor.execute, statement)

        self.failUnless(
            self.TIMEOUT <= time() <= self.TIMEOUT + self.ALLOWED_SLOP,
            "Wallclock duration of execute() call out of bounds.")

        self.assertEquals(timeoutException.statement, statement)
        self.assertEquals(timeoutException.timeout, self.TIMEOUT)
        self.failUnless(isinstance(
                timeoutException.underlying,
                self.expectedUnderlyingExceptionClass))





# This module is really a placeholder: inheritance between database classes is
# unsupported in XAtop right now.  We are just making sure that it is
# aggressively unsupported.

from twisted.trial import unittest

from axiom.item import Item, NoInheritance
from axiom.attributes import integer

class InheritanceUnsupported(unittest.TestCase):

    def testNoInheritance(self):
        class XA(Item):
            schemaVersion = 1
            typeName = 'inheritance_test_xa'
            a = integer()

        try:
            class XB(XA):
                schemaVersion = 1
                typeName = 'inheritance_test_xb'
                b = integer()
        except NoInheritance:
            pass
        else:
            self.fail("Expected RuntimeError but none occurred")

